/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T), Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotStateComponent::
 * @author     ( stefan dot ulbrich at kit dot edu)
 * @date
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include "SharedRobotServants.h"

#include <ArmarXCore/core/Component.h>
#include <ArmarXCore/core/application/properties/Properties.h>
#include <ArmarXCore/core/system/ImportExportComponent.h>
#include <RobotAPI/interface/core/RobotState.h>
#include "SharedRobotServants.h"

#include <RobotAPI/libraries/core/RobotAPIObjectFactories.h>

#include <RobotAPI/libraries/core/remoterobot/RobotStateObserver.h>

#include <ArmarXCore/core/rapidxml/wrapper/RapidXmlReader.h>

namespace armarx
{
    /**
     * \class RobotStatePropertyDefinition
     * \brief
     */
    class RobotStatePropertyDefinitions:
        public ComponentPropertyDefinitions
    {
    public:
        RobotStatePropertyDefinitions(std::string prefix):
            ComponentPropertyDefinitions(prefix)
        {
            defineRequiredProperty<std::string>("RobotNodeSetName", "Set of nodes that is controlled by the KinematicUnit");
            defineRequiredProperty<std::string>("RobotFileName", "Filename of VirtualRobot robot model (e.g. robot_model.xml)");
            defineRequiredProperty<std::string>("AgentName", "Name of the agent for which the sensor values are provided");
            defineOptionalProperty<std::string>("RobotStateReportingTopic", "RobotStateUpdates", "Name of the topic on which updates of the robot state are reported.");
            defineOptionalProperty<int>("HistoryLength", 10000, "Number of entries in the robot state history");
            defineOptionalProperty<float>("RobotModelScaling", 1.0f, "Scaling of the robot model");
        }
    };


    /**
     * \defgroup Component-RobotStateComponent RobotStateComponent
     * \ingroup RobotAPI-Components
     * \brief Maintains a robot representation based on VirtualRobot (see [Simox](http://gitlab.com/Simox/simox)).
     *
     * The robot can be loaded from a Simox robot XML file.
     * Upon request, an Ice proxy to a shared instance of this internal robot
     * can be acquired through a call to RobotStateComponent::getSynchronizedRobot().
     * Additionally it is possible to retrieve a proxy for robot snapshot with
     * RobotStateComponent::getRobotSnapshot().
     * While the synchronized robot will constantly update its internal state
     * the robot snapshot is a clone of the original robot and won't update its
     * configuration over time.  This is useful, if several components need
     * to calculate on the same values.
     * See \ref armarx::RemoteRobot "RemoteRobot" for more details and the usage of this component.
     */

    /**
     * @brief The RobotStateComponent class
     * @ingroup Component-RobotStateComponent
     */
    class ARMARXCOMPONENT_IMPORT_EXPORT RobotStateComponent :
        virtual public Component,
        virtual public RobotStateComponentInterface
    {
    public:

        /**
         * \return SharedRobotInterface proxy to the internal synchronized robot.
         */
        SharedRobotInterfacePrx getSynchronizedRobot(const Ice::Current&) const override;

        /**
         * Creates a snapshot of the robot state at this moment in time.
         * \note You need to call destroy() or ref()/unref() to trigger deletion of this object.
         * \return Clone of the internal synchronized robot fixed to the configuration from the time of calling this function.
         */
        SharedRobotInterfacePrx getRobotSnapshot(const std::string& time, const Ice::Current&) override;

        SharedRobotInterfacePrx getRobotSnapshotAtTimestamp(double time, const Ice::Current& current) override;
        NameValueMap getJointConfigAtTimestamp(double, const Ice::Current&) const override;
        RobotStateConfig getRobotStateAtTimestamp(double timestamp, const Ice::Current&) const override;

        /**
         * \return the robot xml filename as specified in the configuration
         */
        std::string getRobotFilename(const Ice::Current&) const override;

        /*!
         * \brief getArmarXPackages
         * \return All dependent packages, which might contain a robot file.
         */
        std::vector< std::string > getArmarXPackages(const Ice::Current&) const override;

        /**
         *
         * \return  The name of this robot instance.
         */
        std::string getRobotName(const Ice::Current&) const override;

        std::string getRobotStateTopicName(const Ice::Current&) const override;

        /**
         *
         * \return  The name of this robot instance.
         */
        std::string getRobotNodeSetName(const Ice::Current&) const override;

        /**
         * Create an instance of RobotStatePropertyDefinitions.
         */
        PropertyDefinitionsPtr createPropertyDefinitions() override;

        std::string getDefaultName() const override
        {
            return "RobotStateComponent";
        }
        void setRobotStateObserver(RobotStateObserverPtr observer);

        bool interpolate(double time, RobotStateConfig& config) const;

        float getScaling(const Ice::Current&) const override;

        RobotInfoNodePtr getRobotInfo(const Ice::Current&) const override;

    protected:
        /**
         * Load and create a VirtualRobot::Robot instance from the RobotFileName
         * property. Additionally listen on the KinematicUnit topic for the
         * RobotNodeSet specified in the RobotNodeSetName property.
         */
        void onInitComponent() override;
        /**
         * Setup RobotStateObjectFactories needed for creating RemoteRobot instances.
         */
        void onConnectComponent() override;
        void onDisconnectComponent() override;
        /**
         * Calls unref() on RobotStateComponent::_synchronizedPrx.
         */
        ~RobotStateComponent() override;

        // inherited from KinematicUnitInterface
        void reportControlModeChanged(const NameControlModeMap& jointModes, Ice::Long timestamp, bool aValueChanged, const Ice::Current& c = ::Ice::Current()) override;
        void reportJointAngles(const NameValueMap& jointAngles, Ice::Long timestamp,  bool aValueChanged, const Ice::Current& c = ::Ice::Current()) override;
        void reportJointVelocities(const NameValueMap& jointVelocities, Ice::Long timestamp,  bool aValueChanged, const Ice::Current& c = ::Ice::Current()) override;
        void reportJointTorques(const NameValueMap& jointTorques, Ice::Long timestamp,  bool aValueChanged, const Ice::Current& c = ::Ice::Current()) override;
        void reportJointAccelerations(const NameValueMap& jointAccelerations, Ice::Long timestamp, bool aValueChanged, const Ice::Current& c) override;
        void reportJointCurrents(const NameValueMap& jointCurrents, Ice::Long timestamp,  bool aValueChanged, const Ice::Current& c = ::Ice::Current()) override;
        void reportJointMotorTemperatures(const NameValueMap& jointMotorTemperatures, Ice::Long timestamp,  bool aValueChanged, const Ice::Current& c = ::Ice::Current()) override;
        void reportJointStatuses(const NameStatusMap& jointStatuses, Ice::Long timestamp,  bool aValueChanged, const Ice::Current& c = ::Ice::Current()) override;

    private:
        void readRobotInfo(const std::string& robotFile);
        RobotInfoNodePtr readRobotInfo(const RapidXmlReaderNode& infoNode);

        SharedRobotInterfacePrx _synchronizedPrx;
        SharedRobotServantPtr _sharedRobotServant;
        VirtualRobot::RobotPtr _synchronized;
        RobotStateListenerInterfacePrx robotStateListenerPrx;
        std::string robotStateTopicName;
        std::string robotFile;
        std::string relativeRobotFile;
        RobotStateObserverPtr robotStateObs;

        mutable SharedMutex historyMutex;
        std::map<double, RobotStateConfig> history;
        size_t historyLength;

        std::string robotNodeSetName;

        float robotModelScaling;

        RobotInfoNodePtr robotInfo;
    };

}

