/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::NJointController
 * @author     Raphael ( ufdrv at student dot kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#ifndef _ARMARX_LIB_RobotAPI_NJointController_HPP
#define _ARMARX_LIB_RobotAPI_NJointController_HPP

#include <thread>

#include <ArmarXCore/core/util/TemplateMetaProgramming.h>
#include <ArmarXCore/core/util/OnScopeExit.h>

#include "NJointController.h"
#include "../RobotUnit.h"

namespace armarx
{
    namespace RobotUnitModule
    {
        class ControllerManagement;
    }

    class NJointControllerRegistryEntry
    {
    private:
        friend class RobotUnitModule::ControllerManagement;
        virtual NJointControllerPtr create(
                RobotUnitModule::ControllerManagement* cmngr,
                const NJointControllerConfigPtr&,
                const boost::shared_ptr<VirtualRobot::Robot>&,
                bool deletable,
                bool internal) const = 0;
        virtual WidgetDescription::WidgetPtr GenerateConfigDescription(const boost::shared_ptr<VirtualRobot::Robot>&, const std::map<std::string, ConstControlDevicePtr>&, const std::map<std::string, ConstSensorDevicePtr>&) const = 0;
        virtual NJointControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap&) const = 0;
        virtual bool hasRemoteConfiguration() const = 0;
    protected:
        static thread_local bool ConstructorIsRunning_;
    public:
        static bool ConstructorIsRunning()
        {
            return ConstructorIsRunning_;
        }
    };
    using NJointControllerRegistry = Registrar<std::unique_ptr<NJointControllerRegistryEntry>>;

    template<class ControllerType>
    struct NJointControllerRegistration;

    template <typename ControlDataStruct>
    class NJointControllerWithTripleBuffer: virtual public NJointController
    {
    public:
        using MutexType = std::recursive_mutex;
        using LockGuardType = std::lock_guard<std::recursive_mutex>;

        NJointControllerWithTripleBuffer(const ControlDataStruct& initialCommands = ControlDataStruct()):
            controlDataTripleBuffer {initialCommands}
        {
        }

        void rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration) override;

    protected:
        const ControlDataStruct& rtGetControlStruct() const;
        bool                     rtUpdateControlStruct();

        void writeControlStruct();
        ControlDataStruct& getWriterControlStruct();

        void setControlStruct(const ControlDataStruct& newStruct);

        void reinitTripleBuffer(const ControlDataStruct& initial);

        mutable MutexType controlDataMutex;
    private:
        WriteBufferedTripleBuffer<ControlDataStruct> controlDataTripleBuffer;
    };

    template <typename ControlDataStruct>
    using NJointControllerWithTripleBufferPtr = IceInternal::Handle<NJointControllerWithTripleBuffer<ControlDataStruct>>;
}

//inline functions
namespace armarx
{
    template<class T>
    inline T* NJointController::useControlTarget(const std::string& deviceName, const std::string& controlMode)
    {
        static_assert(std::is_base_of<ControlTargetBase, T>::value, "The given type does not derive ControlTargetBase");
        ControlTargetBase* const ptr = useControlTarget(deviceName, controlMode);
        return ptr ? ptr->asA<T>(): nullptr;
    }
    template<class T>
    inline const T* NJointController::useSensorValue(const std::string& deviceName) const
    {
        static_assert(std::is_base_of<SensorValueBase, T>::value, "The given type does not derive SensorValueBase");
        const SensorValueBase* const ptr = useSensorValue(deviceName);
        return ptr ? ptr->asA<T>(): nullptr;
    }

    inline void NJointController::rtSwapBufferAndRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        rtRun(sensorValuesTimestamp, timeSinceLastIteration);
    }

    inline bool NJointController::rtUsesControlDevice(std::size_t deviceIndex) const
    {
        return controlDeviceUsedBitmap.at(deviceIndex);
    }

    inline StringStringDictionary NJointController::getControlDeviceUsedControlModeMap(const Ice::Current&) const
    {
        return controlDeviceControlModeMap;
    }

    inline const std::vector<char>& NJointController::getControlDeviceUsedBitmap() const
    {
        return controlDeviceUsedBitmap;
    }

    inline const std::vector<std::size_t>& NJointController::rtGetControlDeviceUsedIndices() const
    {
        return controlDeviceUsedIndices;
    }

    inline const std::vector<std::size_t>& NJointController::getControlDeviceUsedIndices() const
    {
        return controlDeviceUsedIndices;
    }

    inline const std::map<std::string, const JointController*>& NJointController::getControlDevicesUsedJointController()
    {
        return controlDeviceUsedJointController;
    }

    inline boost::optional<std::vector<char> > NJointController::isNotInConflictWith(const NJointControllerPtr& other) const
    {
        return isNotInConflictWith(other->getControlDeviceUsedBitmap());
    }

    inline std::string NJointController::getDefaultName() const
    {
        return getClassName();
    }

    inline bool NJointController::isControllerActive(const Ice::Current&) const
    {
        return isActive;
    }

    inline bool NJointController::isControllerRequested(const Ice::Current&) const
    {
        return isRequested;
    }

    inline bool NJointController::isDeletable(const Ice::Current&) const
    {
        return deletable;
    }

    inline bool NJointController::hasControllerError(const Ice::Current&) const
    {
        return deactivatedBecauseOfError;
    }

    inline std::string NJointController::getInstanceName(const Ice::Current&) const
    {
        return getName();
    }

    inline WidgetDescription::StringWidgetDictionary NJointController::getFunctionDescriptions(const Ice::Current&) const
    {
        return {};
    }

    inline void NJointController::callDescribedFunction(const std::string&, const StringVariantBaseMap&, const Ice::Current&)
    {
    }

    inline void NJointController::rtSetErrorState()
    {
        errorState.store(true);
    }

    inline bool NJointController::rtGetErrorState() const
    {
        return errorState;
    }

    inline std::size_t NJointController::rtGetNumberOfUsedControlDevices() const
    {
        return controlDeviceUsedIndices.size();
    }

    //NJointControllerWithTripleBuffer<ControlDataStruct>
    template <typename ControlDataStruct>
    inline void NJointControllerWithTripleBuffer<ControlDataStruct>::rtSwapBufferAndRun(
        const IceUtil::Time& sensorValuesTimestamp,
        const IceUtil::Time& timeSinceLastIteration)
    {
        rtUpdateControlStruct();
        rtRun(sensorValuesTimestamp, timeSinceLastIteration);
    }

    template <typename ControlDataStruct>
    inline const ControlDataStruct& NJointControllerWithTripleBuffer<ControlDataStruct>::rtGetControlStruct() const
    {
        return controlDataTripleBuffer.getReadBuffer();
    }

    template <typename ControlDataStruct>
    inline bool NJointControllerWithTripleBuffer<ControlDataStruct>::rtUpdateControlStruct()
    {
        return controlDataTripleBuffer.updateReadBuffer();
    }

    template <typename ControlDataStruct>
    inline void NJointControllerWithTripleBuffer<ControlDataStruct>::writeControlStruct()
    {
        //just lock to be save and reduce the impact of an error
        //also this allows code to unlock the mutex before calling this function
        //(can happen if some lockguard in a scope is used)
        LockGuardType lock(controlDataMutex);
        controlDataTripleBuffer.commitWrite();
    }

    template <typename ControlDataStruct>
    inline ControlDataStruct& NJointControllerWithTripleBuffer<ControlDataStruct>::getWriterControlStruct()
    {
        return controlDataTripleBuffer.getWriteBuffer();
    }

    template <typename ControlDataStruct>
    inline void NJointControllerWithTripleBuffer<ControlDataStruct>::setControlStruct(const ControlDataStruct& newStruct)
    {
        LockGuardType lock {controlDataMutex};
        getWriterControlStruct() = newStruct;
        writeControlStruct();
    }

    template <typename ControlDataStruct>
    inline void NJointControllerWithTripleBuffer<ControlDataStruct>::reinitTripleBuffer(const ControlDataStruct& initial)
    {
        controlDataTripleBuffer.reinitAllBuffers(initial);
    }
}

namespace armarx
{
    template<class ItT>
    inline boost::optional<std::vector<char>> NJointController::AreNotInConflict(ItT first, ItT last)
    {
        if (first == last)
        {
            return {true, std::vector<char>{}};
        }
        std::size_t n = (*first)->getControlDeviceUsedBitmap().size();
        std::vector<char> inuse(n, false);
        while (first != last)
        {
            auto r = (*first)->isNotInConflictWith(inuse);
            if (!r)
            {
                return r;
            }
            inuse = std::move(r.get());
            ++first;
        }
        return {true, std::move(inuse)};
    }

    namespace detail
    {
        template<class NJointControllerT>
        class NJointControllerRegistryEntryHelperBase : public NJointControllerRegistryEntry
        {
            NJointControllerPtr create(RobotUnitModule::ControllerManagement* cmngr,
                const NJointControllerConfigPtr& config,
                const boost::shared_ptr<VirtualRobot::Robot>& rob,
                bool deletable,
                bool internal) const final override
            {
                ARMARX_CHECK_EXPRESSION_W_HINT(cmngr, "ControllerManagement module is NULL!");

               using ConfigPtrT = typename NJointControllerT::ConfigPtrT;
               ConfigPtrT cfg = ConfigPtrT::dynamicCast(config);
               ARMARX_CHECK_EXPRESSION_W_HINT(cfg, "The configuration is of the wrong type! it has to be an instance of: "
                                                   << GetTypeString<ConfigPtrT>());

                ARMARX_CHECK_EXPRESSION_W_HINT(!ConstructorIsRunning(), "Two NJointControllers are created at the same time");
                NJointControllerPtr ptr;
                {
                    ConstructorIsRunning_ = true;
                    ARMARX_ON_SCOPE_EXIT{ConstructorIsRunning_ = false;};
                    ptr = new NJointControllerT(cmngr->_modulePtr<RobotUnit>(), cfg, rob);
                }
                ptr->deletable = deletable;
                ptr->internal = internal;
                return ptr;
            }
        };


        template<class NJointControllerT, class = void, class = void>
        class NJointControllerRegistryEntryHelper : public NJointControllerRegistryEntryHelperBase<NJointControllerT>
        {
            WidgetDescription::WidgetPtr GenerateConfigDescription(const boost::shared_ptr<VirtualRobot::Robot>&,
                    const std::map<std::string, ConstControlDevicePtr>&,
                    const std::map<std::string, ConstSensorDevicePtr>&) const final override
            {
                ARMARX_CHECK_EXPRESSION(!"This function should never be called");
            }
            NJointControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap&) const override
            {
                ARMARX_CHECK_EXPRESSION(!"This function should never be called");
            }
            bool hasRemoteConfiguration() const override
            {
                return false;
            }
        };

        ARMARX_META_MAKE_HAS_MEMBER_FNC_CHECK(hasGenerateConfigDescription,  GenerateConfigDescription,
                                              NJointController::GenerateConfigDescriptionFunctionSignature);
        ARMARX_META_MAKE_HAS_MEMBER_FNC_CHECK(hasGenerateConfigFromVariants,  GenerateConfigFromVariants,
                                              NJointController::GenerateConfigFromVariantsFunctionSignature<typename T::ConfigPtrT>);
#define ARMARX_ASSERT_NJOINTCONTROLLER_HAS_CONSTRUCTION_GUI(T)                                                \
    static_assert(::armarx::detail::hasGenerateConfigDescription<T>::value ,                                \
                  #T " does not offer a construction gui (missing: static GenerateConfigDescription)" );    \
    static_assert(::armarx::detail::hasGenerateConfigFromVariants<T>::value,                                \
                  #T " does not offer a construction gui (missing: static GenerateConfigFromVariants)")

        template<class NJointControllerT>
        class NJointControllerRegistryEntryHelper <
            NJointControllerT,
            typename std::enable_if < hasGenerateConfigDescription<NJointControllerT>::value >::type,
            typename std::enable_if < hasGenerateConfigFromVariants<NJointControllerT>::value >::type
            > : public NJointControllerRegistryEntryHelperBase<NJointControllerT>
        {
            WidgetDescription::WidgetPtr GenerateConfigDescription(
                const boost::shared_ptr<VirtualRobot::Robot>& robot,
                const std::map<std::string, ConstControlDevicePtr>& controlDevices,
                const std::map<std::string, ConstSensorDevicePtr>& sensorDevices
            ) const final override
            {
                try
                {
                    return NJointControllerT::GenerateConfigDescription(robot, controlDevices, sensorDevices);
                }
                catch (Ice::UserException& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigDescription'"
                                 << "\n---- file = " << e.ice_file()
                                 << "\n---- line = " << e.ice_line()
                                 << "\n---- name = " << e.ice_name()
                                 << "\n---- what:\n" << e.what()
                                 << "\n---- stacktrace:\n" << e.ice_stackTrace();
                    throw;
                }
                catch (std::exception& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigDescription'"
                                 << "\n---- what:\n" << e.what();
                    throw;
                }
                catch (...)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigDescription'";
                    throw;
                }
            }
            NJointControllerConfigPtr GenerateConfigFromVariants(const StringVariantBaseMap& variants) const override
            {
                try
                {
                    return NJointControllerT::GenerateConfigFromVariants(variants);
                }
                catch (Ice::UserException& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigFromVariants'"
                                 << "\n---- file = " << e.ice_file()
                                 << "\n---- line = " << e.ice_line()
                                 << "\n---- name = " << e.ice_name()
                                 << "\n---- what:\n" << e.what()
                                 << "\n---- stacktrace:\n" << e.ice_stackTrace();
                    throw;
                }
                catch (std::exception& e)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigFromVariants'"
                                 << "\n---- what:\n" << e.what();
                    throw;
                }
                catch (...)
                {
                    ARMARX_ERROR << "Exception calling '" << GetTypeString<NJointControllerT>() << "::GenerateConfigFromVariants'";
                    throw;
                }
            }
            bool hasRemoteConfiguration() const override
            {
                return true;
            }
        };
    }

    using NJointControllerRegistry = Registrar<std::unique_ptr<NJointControllerRegistryEntry>>;
    template<class ControllerType>
    struct NJointControllerRegistration
    {
        NJointControllerRegistration(const std::string& name)
        {
            NJointControllerRegistry::registerElement(
                name,
                std::unique_ptr<NJointControllerRegistryEntry>(new detail::NJointControllerRegistryEntryHelper<ControllerType>));
        }
    };
}


#endif
