/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    TaskSpaceActiveImpedanceControl::ArmarXObjects::NJointTaskSpaceImpedanceController
 * @author     zhou ( you dot zhou at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "NJointTaskSpaceImpedanceController.h"


using namespace armarx;

NJointControllerRegistration<NJointTaskSpaceImpedanceController> registrationControllerDSRTController("NJointTaskSpaceImpedanceController");


NJointTaskSpaceImpedanceController::NJointTaskSpaceImpedanceController(const RobotUnitPtr& robotUnit, const NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&)
{

    NJointTaskSpaceImpedanceControlConfigPtr cfg = NJointTaskSpaceImpedanceControlConfigPtr::dynamicCast(config);

    ARMARX_CHECK_EXPRESSION(cfg);
    ARMARX_CHECK_EXPRESSION(robotUnit);
    ARMARX_CHECK_EXPRESSION(!cfg->nodeSetName.empty());
    useSynchronizedRtRobot();
    VirtualRobot::RobotNodeSetPtr rns = rtGetRobot()->getRobotNodeSet(cfg->nodeSetName);

    jointNames.clear();
    ARMARX_CHECK_EXPRESSION_W_HINT(rns, cfg->nodeSetName);
    for (size_t i = 0; i < rns->getSize(); ++i)
    {
        std::string jointName = rns->getNode(i)->getName();
        jointNames.push_back(jointName);
        ControlTargetBase* ct = useControlTarget(jointName, ControlModes::Torque1DoF);
        ARMARX_CHECK_EXPRESSION(ct);
        const SensorValueBase* sv = useSensorValue(jointName);
        ARMARX_CHECK_EXPRESSION(sv);
        auto casted_ct = ct->asA<ControlTarget1DoFActuatorTorque>();
        ARMARX_CHECK_EXPRESSION(casted_ct);
        targets.push_back(casted_ct);

        const SensorValue1DoFActuatorTorque* torqueSensor = sv->asA<SensorValue1DoFActuatorTorque>();
        const SensorValue1DoFActuatorVelocity* velocitySensor = sv->asA<SensorValue1DoFActuatorVelocity>();
        const SensorValue1DoFActuatorPosition* positionSensor = sv->asA<SensorValue1DoFActuatorPosition>();
        if (!torqueSensor)
        {
            ARMARX_WARNING << "No Torque sensor available for " << jointName;
        }


        torqueSensors.push_back(torqueSensor);
        velocitySensors.push_back(velocitySensor);
        positionSensors.push_back(positionSensor);
    };
    tcp = rns->getTCP();
    ik.reset(new VirtualRobot::DifferentialIK(rns, rns->getRobot()->getRootNode(), VirtualRobot::JacobiProvider::eSVDDamped));

    ARMARX_CHECK_EQUAL(cfg->desiredPose.size(), 7);
    desiredPosition(0) = cfg->desiredPose[0];
    desiredPosition(1) = cfg->desiredPose[1];
    desiredPosition(2) = cfg->desiredPose[2];

    desiredQuaternion.w() = cfg->desiredPose[3];
    desiredQuaternion.x() = cfg->desiredPose[4];
    desiredQuaternion.y() = cfg->desiredPose[5];
    desiredQuaternion.z() = cfg->desiredPose[6];
    //    desiredTwist = cfg->desiredTwist;
    //    ARMARX_CHECK_EQUAL(desiredPose.rows(), 6);


    ARMARX_CHECK_EQUAL(cfg->Kpos.size(), 3);
    ARMARX_CHECK_EQUAL(cfg->Dpos.size(), 3);
    ARMARX_CHECK_EQUAL(cfg->Kori.size(), 3);
    ARMARX_CHECK_EQUAL(cfg->Dori.size(), 3);

    kpos << cfg->Kpos[0], cfg->Kpos[1], cfg->Kpos[2];
    dpos << cfg->Dpos[0], cfg->Dpos[1], cfg->Dpos[2];
    kori << cfg->Kori[0], cfg->Kori[1], cfg->Kori[2];
    dori << cfg->Dori[0], cfg->Dori[1], cfg->Dori[2];

    torqueLimit = cfg->torqueLimit;

    desiredJointPosition.resize(targets.size());

    ARMARX_CHECK_EQUAL(cfg->desiredJointPositions.size(), targets.size());

    for (size_t i = 0; i < targets.size(); ++i)
    {
        desiredJointPosition(i) = cfg->desiredJointPositions.at(i);
    }


    knull = cfg->Knull;
    dnull = cfg->Dnull;

    NJointTaskSpaceImpedanceControllerControlData initData;
    initData.desiredPose.resize(7);
    initData.desiredPose.setZero();
    initData.desiredTwist.resize(6);
    initData.desiredTwist.setZero();

    reinitTripleBuffer(initData);
}

void NJointTaskSpaceImpedanceController::rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
{
    Eigen::MatrixXf jacobi = ik->getJacobianMatrix(tcp, VirtualRobot::IKSolver::CartesianSelection::All);

    Eigen::VectorXf qpos;
    Eigen::VectorXf qvel;
    qpos.resize(positionSensors.size());
    qvel.resize(velocitySensors.size());

    int jointDim = positionSensors.size();

    for (size_t i = 0; i < velocitySensors.size(); ++i)
    {
        qpos(i) = positionSensors[i]->position;
        qvel(i) = velocitySensors[i]->velocity;
    }

    Eigen::VectorXf tcptwist = jacobi * qvel;

    Eigen::Vector3f currentTCPLinearVelocity;
    currentTCPLinearVelocity << 0.001 * tcptwist(0),   0.001 * tcptwist(1), 0.001 * tcptwist(2);
    Eigen::Vector3f currentTCPAngularVelocity;
    currentTCPAngularVelocity << tcptwist(3),   tcptwist(4),  tcptwist(5);

    Eigen::Vector3f currentTCPPosition = tcp->getPositionInRootFrame();
    Eigen::Vector3f tcpForces = 0.001 * kpos.cwiseProduct(desiredPosition - currentTCPPosition);
    Eigen::Vector3f tcpDesiredForce = tcpForces - dpos.cwiseProduct(currentTCPLinearVelocity);

    Eigen::Matrix4f currentTCPPose = tcp->getPoseInRootFrame();
    Eigen::Matrix3f currentRotMat = currentTCPPose.block<3, 3>(0, 0);
    Eigen::Matrix3f desiredMat(desiredQuaternion);
    Eigen::Matrix3f diffMat = desiredMat * currentRotMat.inverse();
    Eigen::Quaternionf cquat(currentRotMat);
    //    Eigen::Quaternionf diffQuaternion = desiredQuaternion * cquat.conjugate();
    //    Eigen::AngleAxisf angleAxis(diffQuaternion);

    Eigen::Vector3f rpy = VirtualRobot::MathTools::eigen3f2rpy(diffMat);

    Eigen::Vector3f tcpDesiredTorque = kori.cwiseProduct(rpy) - dori.cwiseProduct(currentTCPAngularVelocity);
    Eigen::Vector6f tcpDesiredWrench;
    tcpDesiredWrench <<   0.001 * tcpDesiredForce, tcpDesiredTorque;

    // calculate desired joint torque
    Eigen::MatrixXf I = Eigen::MatrixXf::Identity(jointDim, jointDim);

    float lambda = 2;

    Eigen::MatrixXf jtpinv = ik->computePseudoInverseJacobianMatrix(jacobi.transpose(), lambda);

    Eigen::VectorXf nullqerror = desiredJointPosition - qpos;

    for (int i = 0; i < nullqerror.rows(); ++i)
    {
        if (fabs(nullqerror(i)) < 0.09)
        {
            nullqerror(i) = 0;
        }
    }
    Eigen::VectorXf nullspaceTorque = knull * nullqerror - dnull * qvel;

    Eigen::VectorXf jointDesiredTorques = jacobi.transpose() * tcpDesiredWrench + (I - jacobi.transpose() * jtpinv) * nullspaceTorque;

    for (size_t i = 0; i < targets.size(); ++i)
    {
        float desiredTorque = jointDesiredTorques(i);

        desiredTorque = (desiredTorque >  torqueLimit) ? torqueLimit : desiredTorque;
        desiredTorque = (desiredTorque < -torqueLimit) ? -torqueLimit : desiredTorque;

        debugDataInfo.getWriteBuffer().desired_torques[jointNames[i]] = jointDesiredTorques(i);

        targets.at(i)->torque = desiredTorque;
    }


    float errorAngle = cquat.angularDistance(desiredQuaternion);
    float posiError = (desiredPosition - currentTCPPosition).norm();

    debugDataInfo.getWriteBuffer().desiredForce_x = tcpDesiredForce(0);
    debugDataInfo.getWriteBuffer().desiredForce_y = tcpDesiredForce(1);
    debugDataInfo.getWriteBuffer().desiredForce_z = tcpDesiredForce(2);
    debugDataInfo.getWriteBuffer().tcpDesiredTorque_x = tcpDesiredTorque(0);
    debugDataInfo.getWriteBuffer().tcpDesiredTorque_y = tcpDesiredTorque(1);
    debugDataInfo.getWriteBuffer().tcpDesiredTorque_z = tcpDesiredTorque(2);
    debugDataInfo.getWriteBuffer().quatError = errorAngle;
    debugDataInfo.getWriteBuffer().posiError = posiError;
    debugDataInfo.commitWrite();

}

void NJointTaskSpaceImpedanceController::onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx& debugObs)
{
    StringVariantBaseMap datafields;
    auto values = debugDataInfo.getUpToDateReadBuffer().desired_torques;
    for (auto& pair : values)
    {
        datafields[pair.first] = new Variant(pair.second);
    }

    datafields["desiredForce_x"] = new Variant(debugDataInfo.getUpToDateReadBuffer().desiredForce_x);
    datafields["desiredForce_y"] = new Variant(debugDataInfo.getUpToDateReadBuffer().desiredForce_y);
    datafields["desiredForce_z"] = new Variant(debugDataInfo.getUpToDateReadBuffer().desiredForce_z);

    datafields["tcpDesiredTorque_x"] = new Variant(debugDataInfo.getUpToDateReadBuffer().tcpDesiredTorque_x);
    datafields["tcpDesiredTorque_y"] = new Variant(debugDataInfo.getUpToDateReadBuffer().tcpDesiredTorque_y);
    datafields["tcpDesiredTorque_z"] = new Variant(debugDataInfo.getUpToDateReadBuffer().tcpDesiredTorque_z);

    datafields["quatError"] = new Variant(debugDataInfo.getUpToDateReadBuffer().quatError);
    datafields["posiError"] = new Variant(debugDataInfo.getUpToDateReadBuffer().posiError);

    debugObs->setDebugChannel("DSControllerOutput", datafields);


}

void NJointTaskSpaceImpedanceController::setTarget(const Ice::FloatSeq& target, const Ice::Current&)
{
    Eigen::VectorXf desiredPose;
    desiredPose.resize(7);
    for (size_t i = 0; i < 7; ++i)
    {
        desiredPose(i) = target.at(i);
    }

    LockGuardType guard {controlDataMutex};
    getWriterControlStruct().desiredPose = desiredPose;
    writeControlStruct();
}



