/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <VirtualRobot/XML/RobotIO.h>

#include <ArmarXCore/core/util/OnScopeExit.h>

#include "RobotUnitModuleControlThread.h"
#include "../NJointControllers/NJointController.h"

#include "../Devices/RTThreadTimingsSensorDevice.h"

namespace armarx
{
    namespace RobotUnitModule
    {
        /**
         * \brief This class allows minimal access to private members of \ref NJointController in a sane fashion for \ref ControlThread.
         * \warning !! DO NOT ADD ANYTHING IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
         */
        class NJointControllerAttorneyForControlThread
        {
            friend class ControlThread;
            static void RtDeactivateController(const NJointControllerPtr& nJointCtrl)
            {
                nJointCtrl->rtDeactivateController();
            }
            static void RtActivateController(const NJointControllerPtr& nJointCtrl)
            {
                nJointCtrl->rtActivateController();
            }
            static void RtDeactivateControllerBecauseOfError(const NJointControllerPtr& nJointCtrl)
            {
                nJointCtrl->rtDeactivateControllerBecauseOfError();
            }
        };
        /**
         * \brief This class allows minimal access to private members of \ref ControlThreadDataBuffer in a sane fashion for \ref ControlThread.
         * \warning !! DO NOT ADD ANYTHING IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
         */
        class ControlThreadDataBufferAttorneyForControlThread
        {
            friend class ControlThread;
            static std::vector<JointController*>& GetActivatedJointControllers(ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersActivated.getWriteBuffer().jointControllers;
            }
            static std::vector<NJointController*>& GetActivatedNJointControllers(ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersActivated.getWriteBuffer().nJointControllers;
            }
            static std::vector<std::size_t>& GetActivatedJointToNJointControllerAssignement(ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersActivated.getWriteBuffer().jointToNJointControllerAssignement;
            }
            static void AcceptRequestedJointToNJointControllerAssignement(ControlThread* p)
            {
                p->_module<ControlThreadDataBuffer>().controllersActivated.getWriteBuffer().jointToNJointControllerAssignement =
                    p->_module<ControlThreadDataBuffer>().controllersRequested.getReadBuffer().jointToNJointControllerAssignement;
            }
            static void CommitActivatedControllers(ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersActivated.commitWrite();
            }
            static void ResetActivatedControllerAssignement(ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersActivated.getWriteBuffer().resetAssignement();
            }

            static const std::vector<JointController*>& GetRequestedJointControllers(const ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersRequested.getWriteBuffer().jointControllers;
            }
            static const std::vector<NJointController*>& GetRequestedNJointControllers(const ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersRequested.getWriteBuffer().nJointControllers;
            }
            static bool RequestedControllersChanged(const ControlThread* p)
            {
                return p->_module<ControlThreadDataBuffer>().controllersRequested.updateReadBuffer();
            }
        };
        /**
         * \brief This class allows minimal access to private members of \ref Devices in a sane fashion for \ref ControlThread.
         * \warning !! DO NOT ADD ANYTHING IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
         */
        class DevicesAttorneyForControlThread
        {
            friend class ControlThread;
            static const std::vector<ControlDevicePtr>& GetControlDevices(const ControlThread* p)
            {
                return p->_module<Devices>().controlDevices.values();
            }
            static const std::vector<SensorDevicePtr>& GetSensorDevices(const ControlThread* p)
            {
                return p->_module<Devices>().sensorDevices.values();
            }

            static const std::vector<const SensorValueBase*>& GetSensorValues(const ControlThread* p)
            {
                return p->_module<Devices>().sensorValues;
            }
            static const std::vector<std::vector<PropagateConst<ControlTargetBase*>>>& GetControlTargets(const ControlThread* p)
            {
                return p->_module<Devices>().controlTargets;
            }
            static RTThreadTimingsSensorDevice& GetThreadTimingsSensorDevice(const ControlThread* p)
            {
                return *(p->_module<Devices>().rtThreadTimingsSensorDevice);
            }

            static void UpdateRobotWithSensorValues(const ControlThread* p, const VirtualRobot::RobotPtr& robot, const std::vector<VirtualRobot::RobotNodePtr>& robotNodes)
            {
                p->_module<Devices>().updateVirtualRobotFromSensorValues(robot, robotNodes, p->_module<Devices>().sensorValues);
            }
        };
    }
}

namespace armarx
{
    namespace RobotUnitModule
    {
        bool ControlThread::rtSwitchControllerSetup()
        {
            rtGetThreadTimingsSensorDevice().rtMarkRtSwitchControllerSetupStart();
            ARMARX_ON_SCOPE_EXIT { rtGetThreadTimingsSensorDevice().rtMarkRtSwitchControllerSetupEnd(); };
            // !emergencyStop && !rtIsInEmergencyStop() -> normal control flow
            // !emergencyStop &&  rtIsInEmergencyStop() -> force switch to reactivate old ( + reset flag)
            //  emergencyStop && !rtIsInEmergencyStop() -> deactivate all + set flag
            //  emergencyStop &&  rtIsInEmergencyStop() -> nothing to do

            if (emergencyStop)
            {
                if (rtIsInEmergencyStop())
                {
                    return false;
                }
                rtIsInEmergencyStop_ = true;
                bool updatedActive = false;
                //deactivate all nJointCtrl
                for (auto& nJointCtrl : rtGetActivatedNJointControllers())
                {
                    if (nJointCtrl)
                    {
                        NJointControllerAttorneyForControlThread::RtDeactivateController(nJointCtrl);
                        nJointCtrl = nullptr;
                        updatedActive = true;
                    }
                }
                //set all JointCtrl to emergency stop (except stop movement)
                for (std::size_t i = 0; i < rtGetControlDevices().size(); ++i)
                {
                    const ControlDevicePtr& controlDevice = rtGetControlDevices().at(i);
                    const auto active = controlDevice->rtGetActiveJointController();
                    const auto stopMov = controlDevice->rtGetJointStopMovementController();
                    const auto emergency = controlDevice->rtGetJointEmergencyStopController();
                    if (active != stopMov && active != emergency)
                    {
                        controlDevice->rtSetActiveJointController(emergency);
                        rtGetActivatedJointControllers().at(i) = emergency;
                        updatedActive = true;
                    }
                }
                if (updatedActive)
                {
                    ControlThreadDataBufferAttorneyForControlThread::ResetActivatedControllerAssignement(this);
                    rtCommitActivatedControllers();
                }
                return updatedActive;
            }

            if (
                !ControlThreadDataBufferAttorneyForControlThread::RequestedControllersChanged(this) &&
                !rtIsInEmergencyStop()
            )
            {
                return false;
            }
            rtIsInEmergencyStop_ = false;

            //handle nJointCtrl
            {
                //"merge"
                std::size_t n = rtGetControlDevices().size();
                std::size_t idxAct = 0;
                std::size_t idxReq = 0;
                for (std::size_t i = 0; i < 2 * n; ++i)
                {
                    //skip nullptrs in act
                    while (idxAct < n && !rtGetActivatedNJointControllers().at(idxAct))
                    {
                        ++idxAct;
                    }
                    const NJointControllerPtr& req = idxReq < n ? rtGetRequestedNJointControllers().at(idxReq) : nullptr; //may be null
                    const NJointControllerPtr& act = idxAct < n ? rtGetActivatedNJointControllers().at(idxAct) : nullptr; //may be null if it is the last
                    const auto reqId = reinterpret_cast<std::uintptr_t>(req.get());
                    const auto actId = reinterpret_cast<std::uintptr_t>(act.get());

                    if (reqId > actId)
                    {
                        //new ctrl
                        rtSyncNJointControllerRobot(req.get());
                        NJointControllerAttorneyForControlThread::RtActivateController(req);
                        ++idxReq;
                    }
                    else if (reqId < actId)
                    {
                        NJointControllerAttorneyForControlThread::RtDeactivateController(act);
                        ++idxAct;
                    }
                    else //if(reqId == actId)
                    {
                        //same ctrl or both null ctrl
                        ++idxReq;
                        ++idxAct;
                    }
                    if (idxAct >= n && !req)
                    {
                        break;
                    }
                }
                rtGetActivatedNJointControllers() = rtGetRequestedNJointControllers();
            }

            //handle Joint Ctrl
            {
                ARMARX_CHECK_EQUAL(rtGetRequestedJointControllers().size(), rtGetControlDevices().size());
                for (std::size_t i = 0; i < rtGetControlDevices().size(); ++i)
                {
                    auto& controlDevice = rtGetControlDevices().at(i);
                    const auto requestedJointCtrl = rtGetRequestedJointControllers().at(i);
                    controlDevice->rtSetActiveJointController(requestedJointCtrl);
                    rtGetActivatedJointControllers().at(i) = requestedJointCtrl;
                }
            }
            ControlThreadDataBufferAttorneyForControlThread::AcceptRequestedJointToNJointControllerAssignement(this);
            rtCommitActivatedControllers();
            return true;
        }

        void ControlThread::rtResetAllTargets()
        {
            rtGetThreadTimingsSensorDevice().rtMarkRtResetAllTargetsStart();
            for (const ControlDevicePtr& controlDev : rtGetControlDevices())
            {
                controlDev->rtGetActiveJointController()->rtResetTarget();
            }
            rtGetThreadTimingsSensorDevice().rtMarkRtResetAllTargetsEnd();
        }

        void ControlThread::rtHandleInvalidTargets()
        {
            rtGetThreadTimingsSensorDevice().rtMarkRtHandleInvalidTargetsStart();
            bool oneIsInvalid = 0;
            for (std::size_t i = 0; i < rtGetControlDevices().size(); ++i)
            {
                if (!rtGetActivatedJointControllers().at(i)->rtIsTargetVaild())
                {
                    ARMARX_ERROR << "INVALID TARGET: " << rtGetControlDevices().at(i)->getDeviceName();
                    rtDeactivateAssignedNJointControllerBecauseOfError(i, false);
                    oneIsInvalid = true;
                }
            }
            if (oneIsInvalid)
            {
                rtCommitActivatedControllers();
            }
            rtGetThreadTimingsSensorDevice().rtMarkRtHandleInvalidTargetsEnd();
        }

        void ControlThread::rtReadSensorDeviceValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
        {
            rtGetThreadTimingsSensorDevice().rtMarkRtReadSensorDeviceValuesStart();
            for (const SensorDevicePtr& device : rtGetSensorDevices())
            {
                device->rtReadSensorValues(sensorValuesTimestamp, timeSinceLastIteration);
            }
            DevicesAttorneyForControlThread::UpdateRobotWithSensorValues(this, rtRobot, rtRobotNodes);
            rtGetThreadTimingsSensorDevice().rtMarkRtReadSensorDeviceValuesEnd();
        }

        void ControlThread::rtRunJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
        {
            rtGetThreadTimingsSensorDevice().rtMarkRtRunJointControllersStart();
            for (const ControlDevicePtr& device : rtGetControlDevices())
            {
                device->rtRun(sensorValuesTimestamp, timeSinceLastIteration);
            }
            rtGetThreadTimingsSensorDevice().rtMarkRtRunJointControllersEnd();
        }

        void ControlThread::rtRunNJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
        {
            rtGetThreadTimingsSensorDevice().rtMarkRtRunNJointControllersStart();
            bool activeControllersChanged = false;
            for (std::size_t nJointCtrlIndex = 0; nJointCtrlIndex < rtGetActivatedNJointControllers().size(); ++nJointCtrlIndex)
            {
                NJointController* nJointCtrl = rtGetActivatedNJointControllers().at(nJointCtrlIndex);
                try
                {
                    if (nJointCtrl)
                    {
                        if (nJointCtrl->rtGetErrorState())
                        {
                            rtDeactivateNJointControllerBecauseOfError(nJointCtrlIndex, false);
                            activeControllersChanged = true;
                        }

                        auto start = TimeUtil::GetTime(true);
                        rtSyncNJointControllerRobot(nJointCtrl);
                        nJointCtrl->rtSwapBufferAndRun(sensorValuesTimestamp, timeSinceLastIteration);
                        auto duration = TimeUtil::GetTime(true) - start;
                        if (nJointCtrl->rtGetErrorState())
                        {
                            rtDeactivateNJointControllerBecauseOfError(nJointCtrlIndex, false);
                            activeControllersChanged = true;
                        }
                        if (static_cast<std::size_t>(duration.toMicroSeconds()) > nJointCtrl->rtGetNumberOfUsedControlDevices() * usPerDevUntilError)
                        {
                            ARMARX_ERROR << deactivateSpam(5) << "The NJointController " << nJointCtrl->getClassName() << " took " << duration.toMicroSeconds() << " µs to run!";
                        }
                        else if (static_cast<std::size_t>(duration.toMicroSeconds()) > nJointCtrl->rtGetNumberOfUsedControlDevices() * usPerDevUntilWarn)
                        {
                            ARMARX_WARNING << deactivateSpam(5) << "The NJointController " << nJointCtrl->getClassName() << " took " << duration.toMicroSeconds() << " µs to run!";
                        }
                    }
                }
                catch (...)
                {
                    ARMARX_ERROR << "NJoint Controller " << nJointCtrl->getInstanceName()
                                 << " threw an exception and is now deactivated: "
                                 << GetHandledExceptionString();
                    rtDeactivateNJointControllerBecauseOfError(nJointCtrlIndex, false);
                    activeControllersChanged = true;
                }
            }
            if (activeControllersChanged)
            {
                rtCommitActivatedControllers();
            }
            rtGetThreadTimingsSensorDevice().rtMarkRtRunNJointControllersEnd();
        }

        void ControlThread::rtDeactivateNJointControllerBecauseOfError(std::size_t nJointCtrlIndex, bool writeActivatedControlers)
        {
            const NJointControllerPtr& nJointCtrl = rtGetActivatedNJointControllers().at(nJointCtrlIndex);
            NJointControllerAttorneyForControlThread::RtDeactivateControllerBecauseOfError(nJointCtrl);
            for (auto ctrlDevIdx : nJointCtrl->rtGetControlDeviceUsedIndices())
            {
                const ControlDevicePtr& controlDevice = rtGetControlDevices().at(ctrlDevIdx);
                JointController* es = controlDevice->rtGetJointEmergencyStopController();
                controlDevice->rtSetActiveJointController(es);
                rtGetActivatedJointControllers().at(ctrlDevIdx) = es;
                ARMARX_CHECK_EXPRESSION(rtGetActivatedJointToNJointControllerAssignement().at(ctrlDevIdx) == nJointCtrlIndex);
                rtGetActivatedJointToNJointControllerAssignement().at(ctrlDevIdx) = IndexSentinel();
            }
            rtGetActivatedNJointControllers().at(nJointCtrlIndex) = nullptr;
            //check ControlDeviceHardwareControlModeGroups
            {
                for (auto ctrlDevIdx : nJointCtrl->rtGetControlDeviceUsedIndices())
                {
                    const auto& ctrlModeGroups = _module<Devices>().getControlModeGroups();
                    const auto groupIdx = ctrlModeGroups.groupIndices.at(ctrlDevIdx);
                    if (groupIdx == IndexSentinel())
                    {
                        continue;
                    }
                    ControlDevice*  const dev  = rtGetControlDevices().at(ctrlDevIdx).get();
                    JointController* const jointCtrl = dev->rtGetActiveJointController();
                    const auto hwModeHash = jointCtrl->rtGetHardwareControlModeHash();
                    //this device is in a group!
                    // -> check all other devices
                    for (const auto otherIdx : ctrlModeGroups.deviceIndices.at(groupIdx))
                    {
                        ControlDevice*  const otherDev  = rtGetControlDevices().at(otherIdx).get();
                        JointController* const otherJointCtrl = otherDev->rtGetActiveJointController();
                        const auto otherHwModeHash = otherJointCtrl->rtGetHardwareControlModeHash();
                        if (hwModeHash == otherHwModeHash)
                        {
                            //the assigend ctrl has the same hwMode -> don't do anything
                            continue;
                        }
                        const auto otherNJointCtrl1Idx = rtGetActivatedJointToNJointControllerAssignement().at(otherIdx);
                        if (otherNJointCtrl1Idx == IndexSentinel())
                        {
                            //the hwmodes are different! (hence the other ctrl must be in stop movement
                            ARMARX_CHECK_EXPRESSION(otherJointCtrl == otherDev->rtGetJointStopMovementController());
                            //we need to activate the es contoller
                            JointController* const es = otherDev->rtGetJointEmergencyStopController();
                            otherDev->rtSetActiveJointController(es);
                            ARMARX_CHECK_EXPRESSION(es->rtGetHardwareControlModeHash() == hwModeHash);
                            rtGetActivatedJointControllers().at(otherIdx) = es;
                            continue;
                        }
                        //the devs NJoint controller needs to be deactivated
                        rtDeactivateAssignedNJointControllerBecauseOfError(otherIdx, false);
                    }
                }
            }
            rtDeactivatedNJointControllerBecauseOfError(nJointCtrl);

            if (writeActivatedControlers)
            {
                rtCommitActivatedControllers();
            }
        }

        void ControlThread::rtDeactivateAssignedNJointControllerBecauseOfError(std::size_t ctrlDevIndex, bool writeActivatedControlers)
        {
            std::size_t nJointCtrlIndex = rtGetActivatedJointToNJointControllerAssignement().at(ctrlDevIndex);
            ARMARX_CHECK_EXPRESSION_W_HINT(
                nJointCtrlIndex < rtGetControlDevices().size(),
                nJointCtrlIndex << " < " << rtGetControlDevices().size() <<
                ": no NJoint controller controls this device (name = "
                << rtGetControlDevices().at(ctrlDevIndex)->getDeviceName()
                << ", ControlMode = " << rtGetActivatedJointControllers().at(ctrlDevIndex)->getControlMode() << ")!" << "\n"
                << VAROUT(ctrlDevIndex) << "\n"
                << ARMARX_STREAM_PRINTER
            {
                for (std::size_t i = 0; i < rtGetActivatedJointControllers().size(); ++i)
                {
                    const JointController* jctrl = rtGetActivatedJointControllers().at(i);
                    out << "    " << i << "\t" << jctrl << " " << jctrl->getControlMode() << "\n";
                }
            }
                    << VAROUT(rtGetActivatedNJointControllers()) << "\n"
                    << VAROUT(rtGetActivatedJointToNJointControllerAssignement())
            );

            rtDeactivateNJointControllerBecauseOfError(nJointCtrlIndex, false);

            if (writeActivatedControlers)
            {
                rtCommitActivatedControllers();
            }
        }

        void ControlThread::rtUpdateSensorAndControlBuffer(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
        {
            rtGetThreadTimingsSensorDevice().rtMarkRtUpdateSensorAndControlBufferStart();
            SensorAndControl& sc = _module<ControlThreadDataBuffer>().rtGetSensorAndControlBuffer();
            sc.writeTimestamp = IceUtil::Time::now(); // this has to be in real time
            sc.sensorValuesTimestamp = sensorValuesTimestamp;
            sc.timeSinceLastIteration = timeSinceLastIteration;
            ARMARX_CHECK_EQUAL(rtGetSensorDevices().size(), sc.sensors.size());
            for (std::size_t sensIdx = 0; sensIdx < rtGetSensorDevices().size(); ++sensIdx)
            {
                rtGetSensorDevices().at(sensIdx)->getSensorValue()->_copyTo(sc.sensors.at(sensIdx));
            }

            ARMARX_CHECK_EQUAL(rtGetControlDevices().size(), sc.control.size());
            for (std::size_t ctrlIdx = 0; ctrlIdx < rtGetControlDevices().size(); ++ctrlIdx)
            {
                ControlDevice& controlDevice = *rtGetControlDevices().at(ctrlIdx);
                ARMARX_CHECK_EQUAL(controlDevice.rtGetJointControllers().size(), sc.control.at(ctrlIdx).size());
                for (std::size_t targIdx = 0; targIdx < controlDevice.rtGetJointControllers().size(); ++targIdx)
                {
                    JointController& jointCtrl = *controlDevice.rtGetJointControllers().at(targIdx);
                    jointCtrl.getControlTarget()->_copyTo(sc.control.at(ctrlIdx).at(targIdx));
                }
            }
            _module<ControlThreadDataBuffer>().rtSensorAndControlBufferCommitWrite();

            rtGetThreadTimingsSensorDevice().rtMarkRtUpdateSensorAndControlBufferEnd();
        }

        const std::vector<ControlDevicePtr>& ControlThread::rtGetControlDevices()
        {
            return DevicesAttorneyForControlThread::GetControlDevices(this);
        }

        const std::vector<SensorDevicePtr>& ControlThread::rtGetSensorDevices()
        {
            return DevicesAttorneyForControlThread::GetSensorDevices(this);
        }

        RTThreadTimingsSensorDevice& ControlThread::rtGetThreadTimingsSensorDevice()
        {
            return DevicesAttorneyForControlThread::GetThreadTimingsSensorDevice(this);
        }

        void ControlThread::rtSetEmergencyStopState(EmergencyStopState state)
        {
            if (state == EmergencyStopState::eEmergencyStopActive)
            {
                emergencyStopStateRequest = EmergencyStopStateRequest::RequestActive;
            }
            else
            {
                emergencyStopStateRequest = EmergencyStopStateRequest::RequestInactive;
            }
        }

        void ControlThread::_preFinishControlThreadInitialization()
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            controlThreadId = std::this_thread::get_id();
        }

        void ControlThread::_preOnInitRobotUnit()
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            try
            {
                rtRobot = VirtualRobot::RobotIO::loadRobot(_module<RobotData>().getRobotFileName(), VirtualRobot::RobotIO::eStructure);
                rtRobot->setUpdateCollisionModel(false);
                rtRobot->setUpdateVisualization(false);
                rtRobot->setThreadsafe(false);
                rtRobotNodes = rtRobot->getRobotNodes();
            }
            catch (VirtualRobot::VirtualRobotException& e)
            {
                throw UserException(e.what());
            }
            usPerDevUntilWarn = getProperty<std::size_t>("NjointController_AllowedExecutionTimePerControlDeviceUntilWarning").getValue();
            usPerDevUntilError = getProperty<std::size_t>("NjointController_AllowedExecutionTimePerControlDeviceUntilError").getValue();
        }

        void ControlThread::setEmergencyStopState(EmergencyStopState state, const Ice::Current&)
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            _module<Units>().getEmergencyStopMaster()->setEmergencyStopState(state);
        }

        EmergencyStopState ControlThread::getEmergencyStopState(const Ice::Current&) const
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            return emergencyStop ? EmergencyStopState::eEmergencyStopActive : EmergencyStopState::eEmergencyStopInactive;
        }

        EmergencyStopState ControlThread::getRtEmergencyStopState(const Ice::Current&) const
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            return rtIsInEmergencyStop() ? EmergencyStopState::eEmergencyStopActive : EmergencyStopState::eEmergencyStopInactive;
        }

        void ControlThread::setEmergencyStopStateNoReportToTopic(EmergencyStopState state)
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            emergencyStop = (state == EmergencyStopState::eEmergencyStopActive);
        }

        void ControlThread::processEmergencyStopRequest()
        {
            const auto state = emergencyStopStateRequest.exchange(EmergencyStopStateRequest::RequestNone);
            switch (state)
            {
                case EmergencyStopStateRequest::RequestActive:
                    setEmergencyStopState(EmergencyStopState::eEmergencyStopActive);
                    break;
                case EmergencyStopStateRequest::RequestInactive:
                    setEmergencyStopState(EmergencyStopState::eEmergencyStopInactive);
                    break;
                case EmergencyStopStateRequest::RequestNone:
                    break;
                default:
                    ARMARX_CHECK_EXPRESSION_W_HINT(!static_cast<int>(state), "Unkown value for EmergencyStopStateRequest");
            }
        }

        std::vector<JointController*>& ControlThread::rtGetActivatedJointControllers()
        {
            return ControlThreadDataBufferAttorneyForControlThread::GetActivatedJointControllers(this);
        }

        std::vector<NJointController*>& ControlThread::rtGetActivatedNJointControllers()
        {
            return ControlThreadDataBufferAttorneyForControlThread::GetActivatedNJointControllers(this);
        }

        void ControlThread::rtCommitActivatedControllers()
        {
            ControlThreadDataBufferAttorneyForControlThread::CommitActivatedControllers(this);
        }

        const std::vector<JointController*>& ControlThread::rtGetRequestedJointControllers()
        {
            return ControlThreadDataBufferAttorneyForControlThread::GetRequestedJointControllers(this);
        }

        const std::vector<NJointController*>& ControlThread::rtGetRequestedNJointControllers()
        {
            return ControlThreadDataBufferAttorneyForControlThread::GetRequestedNJointControllers(this);
        }

        std::vector<std::size_t>& ControlThread::rtGetActivatedJointToNJointControllerAssignement()
        {
            return ControlThreadDataBufferAttorneyForControlThread::GetActivatedJointToNJointControllerAssignement(this);
        }

        void ControlThread::rtSyncNJointControllerRobot(NJointController* njctrl)
        {
            if (njctrl->rtGetRobot())
            {
                auto& from = rtRobotNodes;
                auto& to = njctrl->rtGetRobotNodes();
                for (std::size_t i = 0; i < from.size(); ++i)
                {
                    to.at(i)->copyPoseFrom(from.at(i));
                }
            }
        }

        void ControlThread::rtSetRobotGlobalPose(const Eigen::Matrix4f& gp, bool updateRobot)
        {
            rtRobot->setGlobalPose(gp, updateRobot);
        }
    }
}
