/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <thread>

#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>

#include "../Devices/ControlDevice.h"
#include "../Devices/SensorDevice.h"

#include "RobotUnitModuleBase.h"

namespace armarx
{
    class RTThreadTimingsSensorDevice;
    namespace RobotUnitModule
    {
        class ControlThreadPropertyDefinitions: public ModuleBasePropertyDefinitions
        {
        public:
            ControlThreadPropertyDefinitions(std::string prefix): ModuleBasePropertyDefinitions(prefix)
            {
                defineOptionalProperty<std::size_t>(
                    "NjointController_AllowedExecutionTimePerControlDeviceUntilWarning", 2,
                    "A Warning will be printed, If the execution time in micro seconds of a NJointController "
                    "exceeds this parameter times the number of ControlDevices."
                );
                defineOptionalProperty<std::size_t>(
                    "NjointController_AllowedExecutionTimePerControlDeviceUntilError", 20,
                    "An Error will be printed, If the execution time in micro seconds of a NJointController "
                    "exceeds this parameter times the number of ControlDevices."
                );
            }
        };

        /**
         * @ingroup Library-RobotUnit-Modules
         * @brief This \ref ModuleBase "Module" manages the \ref ControlThread.
         *
         * It offers the basic functions called in the \ref ControlThread.
         *
         * @see ModuleBase
         */
        class ControlThread : virtual public ModuleBase, virtual public RobotUnitControlThreadInterface
        {
            friend class ModuleBase;
            enum class EmergencyStopStateRequest
            {
                RequestActive,
                RequestInactive,
                RequestNone,
            };
        public:
            /**
             * @brief Returns the singleton instance of this class
             * @return The singleton instance of this class
             */
            static ControlThread& Instance()
            {
                return ModuleBase::Instance<ControlThread>();
            }
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // ///////////////////////////////// RobotUnitModule hooks //////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            ///@see \ref ModuleBase::_preFinishControlThreadInitialization
            void _preFinishControlThreadInitialization();
            ///@see \ref ModuleBase::_preOnInitRobotUnit
            void _preOnInitRobotUnit();
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // ///////////////////////////////////// ice interface //////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        public:
            /**
             * @brief Sets the \ref EmergencyStopState
             * @param state The \ref EmergencyStopState to set
             */
            void setEmergencyStopState(EmergencyStopState state, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Returns the \ref ControlThread's target \ref EmergencyStopState
             * @return The \ref EmergencyStopState
             */
            EmergencyStopState getEmergencyStopState(const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns the \ref ControlThread's \ref EmergencyStopState
             * @return The \ref EmergencyStopState
             */
            EmergencyStopState getRtEmergencyStopState(const Ice::Current& = GlobalIceCurrent) const override;
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // /////////////////////////////////// Module interface /////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        public:
            /**
             * @brief Teturns he \ref ControlThread's thread id
             * @return The \ref ControlThread's thread id
             * @see controlThreadId
             */
            std::thread::id getControlThreadId() const
            {
                return controlThreadId;
            }
        private:
            /**
             * @brief Sets the \ref EmergencyStopState without updating the topic.
             * For use in \ref RobotUnitEmergencyStopMaster.
             * @param state The \ref EmergencyStopState to set
             */
            void setEmergencyStopStateNoReportToTopic(EmergencyStopState state);
            /**
             * @brief Sets the emergency stop state to the request from rt (if there was a request).
             * For use in \ref Publisher.
             */
            void processEmergencyStopRequest();

            // //////////////////////////////////////////////////////////////////////////////////////// //
            // ///////////////////////////////////// rt interface ///////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        protected:
            /**
             * @brief Returns the simox robot used in the control thread
             * @return The simox robot used in the control thread
             */
            const VirtualRobot::RobotPtr& rtGetRobot() const
            {
                return rtRobot;
            }
            /**
             * @brief Changes the set of active controllers.
             * Changes can be caused by a new set of requested controllers or emergency stop
             * @return Whether active controllers were changed
             */
            bool rtSwitchControllerSetup();

            /**
             * @brief rtDeactivateAssignedNJointControllerBecauseOfError
             * @param ctrlDevIndex
             * @param writeActivatedControlers
             */
            void rtDeactivateAssignedNJointControllerBecauseOfError(std::size_t ctrlDevIndex, bool writeActivatedControlers = true);
            /**
             * @brief Deactivates a NJointController and sets all used ControlDevices to EmergencyStop
             * @param nJointCtrlIndex The NJointController to deactivate (index in controllersActivated)
             * @param writeActivatedControlers Whether controllersActivated should be committed
             */
            void rtDeactivateNJointControllerBecauseOfError(std::size_t nJointCtrlIndex, bool writeActivatedControlers = true);
            /**
             * @brief Runs NJoint controllers
             * @param sensorValuesTimestamp Timestamp of the current \ref SensorValueBase SensorValues
             * @param timeSinceLastIteration Time between the last two updates of \ref SensorValueBase SensorValues
             */
            void rtRunNJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
            /**
             * @brief Runs Joint controllers and writes target values to ControlDevices
             * @param sensorValuesTimestamp Timestamp of the current \ref SensorValueBase SensorValues
             * @param timeSinceLastIteration Time between the last two updates of \ref SensorValueBase SensorValues
             */
            void rtRunJointControllers(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
            /**
             * @brief Calls \ref SensorDevice::rtReadSensorValues for all \ref SensorDevice "SensorDevices".
             * @param sensorValuesTimestamp Timestamp of the current \ref SensorValueBase SensorValues
             * @param timeSinceLastIteration Time between the last two updates of \ref SensorValueBase SensorValues
             */
            void rtReadSensorDeviceValues(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);
            /// @brief Deactivates all NJointControllers generating invalid targets.
            void rtHandleInvalidTargets();
            /// @brief Calls \ref JointController::rtResetTarget for all active Joint controllers
            void rtResetAllTargets();
            /**
             * @brief Hook for deriving classes (called AFTER a NJointController is deactivated due to an error)
             *
             * This is useful when two Control devices can be controlled by different NJointControllers but either both or none
             * have to be controlled.
             *
             * Since this hook is called after the controller with an error was deactivated an other call to
             * \ref rtDeactivateAssignedNJointControllerBecauseOfError can't lead to a cyclic execution.
             */
            virtual void rtDeactivatedNJointControllerBecauseOfError(const NJointControllerPtr& /*nJointCtrl*/) {}
            /**
             * @brief Updates the current \ref SensorValueBase "SensorValues" and \ref ControlTargetBase "ControlTargets"
             * for code running outside of the \ref ControlThread
             * @param sensorValuesTimestamp Timestamp of the current \ref SensorValueBase SensorValues
             * @param timeSinceLastIteration Time between the last two updates of \ref SensorValueBase SensorValues
             */
            void rtUpdateSensorAndControlBuffer(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);

            /**
             * @brief Returns all \ref ControlDevice "ControlDevices"
             * @return All \ref ControlDevice "ControlDevices"
             */
            const std::vector<ControlDevicePtr>& rtGetControlDevices();
            /**
             * @brief Returns all \ref SensorDevice "SensorDevices"
             * @return All \ref SensorDevice "SensorDevices"
             */
            const std::vector<SensorDevicePtr>& rtGetSensorDevices();
            /**
             * @brief Returns the \ref RTThreadTimingsSensorDevice
             * @return The \ref RTThreadTimingsSensorDevice
             */
            RTThreadTimingsSensorDevice& rtGetThreadTimingsSensorDevice();

            /**
             * @brief Returns whether the rt thread is in emergency stop. (This function is for use in the \ref ControlThread)
             * @return Whether the rt thread is in emergency stop.
             */
            bool rtIsInEmergencyStop() const
            {
                return rtIsInEmergencyStop_;
            }
            /**
             * @brief Returns the \ref ControlThread's emergency stop state. (This function is for use in the \ref ControlThread)
             * @return The \ref ControlThread's emergency stop state
             */
            EmergencyStopState rtGetEmergencyStopState() const
            {
                return rtIsInEmergencyStop_ ? EmergencyStopState::eEmergencyStopActive : EmergencyStopState::eEmergencyStopInactive;
            }

            /**
             * @brief Set an \ref EmergencyStopState request. This request will be executed by the \ref Publisher
             * @param state The \ref EmergencyStopState to set
             */
            void rtSetEmergencyStopState(EmergencyStopState state);

            void rtSetRobotGlobalPose(const Eigen::Matrix4f& gp, bool updateRobot = true);
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // //////////////////////////////////// implementation //////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            /**
             * @brief Returns the activated \ref JointController "JointControllers"
             * @return The activated \ref JointController "JointControllers"
             */
            std::vector<JointController*>& rtGetActivatedJointControllers();
            /**
             * @brief Returns the activated \ref NJointController "NJointControllers"
             * @return The activated \ref NJointController "NJointControllers"
             */
            std::vector<NJointController*>& rtGetActivatedNJointControllers();
            /**
             * @brief Writes the activated \ref NJointController "NJointControllers" and
             * \ref JointController "JointControllers" to the output buffer.
             */
            void rtCommitActivatedControllers();

            /**
             * @brief Returns the requested \ref JointController "JointControllers"
             * @return The requested \ref JointController "JointControllers"
             */
            const std::vector<JointController*>& rtGetRequestedJointControllers();
            /**
             * @brief Returns the requested \ref NJointController "NJointControllers"
             * @return The requested \ref NJointController "NJointControllers"
             */
            const std::vector<NJointController*>& rtGetRequestedNJointControllers();

            /**
             * @brief Returns a vector containing the index of the activated \ref NJointController
             * for each \ref JointController
             * @return A vector containing the index of the activated \ref NJointController
             * for each \ref JointController
             */
            std::vector<std::size_t>& rtGetActivatedJointToNJointControllerAssignement();

            void rtSyncNJointControllerRobot(NJointController* njctrl);
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////// Data ///////////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            ///@brief Whether the ControlThread is in emergency stop
            std::atomic_bool rtIsInEmergencyStop_ {false};
            ///@brief Whether the ControlThread is supposed to be in emergency stop
            std::atomic_bool emergencyStop {false};

            ///@brief The \ref VirtualRobot::Robot used in the \ref ControlThread
            VirtualRobot::RobotPtr rtRobot;
            ///@brief The \ref VirtualRobot::Robot's \ref VirtualRobot::RobotNode "VirtualRobot::RobotNode" used in the \ref ControlThread
            std::vector<VirtualRobot::RobotNodePtr> rtRobotNodes;

            ///@brief The \ref ControlThread's thread id
            std::atomic<std::thread::id> controlThreadId;

            /// @brief A Warning will be printed, if the execution time per ControlDev of a NJointController exceeds this parameter
            std::size_t usPerDevUntilWarn;
            /// @brief An Error will be printed, if the execution time per ControlDev of a NJointController exceeds this parameter
            std::size_t usPerDevUntilError;

            std::atomic<EmergencyStopStateRequest> emergencyStopStateRequest {EmergencyStopStateRequest::RequestNone};
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // /////////////////////////////////////// Attorneys ////////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            /**
            * \brief This class allows minimal access to private members of \ref ControlThread in a sane fashion for \ref Publisher.
            * \warning !! DO NOT ADD ADDITIONAL FRIENDS IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
            */
            friend class ControlThreadAttorneyForPublisher;
            /**
            * \brief This class allows minimal access to private members of \ref ControlThread in a sane fashion for \ref RobotUnitEmergencyStopMaster.
            * \warning !! DO NOT ADD ADDITIONAL FRIENDS IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
            */
            friend class ControlThreadAttorneyForRobotUnitEmergencyStopMaster;
        };
    }
}
