/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <thread>

#include <RobotAPI/interface/units/RobotUnit/RobotUnitInterface.h>

#include "RobotUnitModuleBase.h"

namespace armarx
{
    namespace RobotUnitModule
    {
        /**
         * @ingroup Library-RobotUnit-Modules
         * @brief This \ref ModuleBase "Module" manages \ref NJointController "NJointControllers".
         *
         * @see ModuleBase
         */
        class ControllerManagement : virtual public ModuleBase, virtual public RobotUnitControllerManagementInterface
        {
            friend class ModuleBase;
        public:
            /**
             * @brief Returns the singleton instance of this class
             * @return The singleton instance of this class
             */
            static ControllerManagement& Instance()
            {
                return ModuleBase::Instance<ControllerManagement>();
            }
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // ///////////////////////////////// RobotUnitModule hooks //////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            /// @see ModuleBase::_preOnInitRobotUnit
            void _preOnInitRobotUnit();
            /// @see ModuleBase::_preFinishRunning
            void _preFinishRunning();
            /// @see ModuleBase::_postFinishRunning
            void _postFinishRunning();
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // ///////////////////////////////////// ice interface //////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        public:
            /**
             * @brief Returns a proxy to the \ref NJointController.
             * @param name The \ref NJointController's name.
             * @return A proxy to the \ref NJointController.
             * @see getAllNJointControllers
             */
            NJointControllerInterfacePrx getNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns proxies to all \ref NJointController "NJointControllers"
             * @return Proxies to all \ref NJointController "NJointControllers"
             * @see getNJointController
             */
            StringNJointControllerPrxDictionary getAllNJointControllers(const Ice::Current& = GlobalIceCurrent) const override;

            /**
             * @brief Returns the status of the \ref NJointController.
             * @param name The \ref NJointController's name.
             * @return The status of the \ref NJointController.
             * @see NJointControllerStatus
             * @see getNJointControllerStatuses
             * @see getNJointControllerDescriptionWithStatus
             * @see getNJointControllerDescriptionsWithStatuses
             */
            NJointControllerStatus getNJointControllerStatus(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns the status of all \ref NJointController "NJointControllers".
             * @return The status of all \ref NJointController "NJointControllers".
             * @see NJointControllerStatus
             * @see getNJointControllerStatus
             * @see getNJointControllerDescriptionWithStatus
             * @see getNJointControllerDescriptionsWithStatuses
             */
            NJointControllerStatusSeq getNJointControllerStatuses(const Ice::Current& = GlobalIceCurrent) const override;

            /**
             * @brief Returns the description of the \ref NJointController.
             * @param name The \ref NJointController's name.
             * @return The description of the \ref NJointController.
             * @see NJointControllerDescription
             * @see getNJointControllerDescriptions
             * @see getNJointControllerDescriptionWithStatus
             * @see getNJointControllerDescriptionsWithStatuses
             */
            NJointControllerDescription getNJointControllerDescription(const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns the description of all \ref NJointController "NJointControllers".
             * @return The description of all \ref NJointController "NJointControllers".
             * @see NJointControllerDescription
             * @see getNJointControllerDescription
             * @see getNJointControllerDescriptionWithStatus
             * @see getNJointControllerDescriptionsWithStatuses
             */

            NJointControllerDescriptionSeq getNJointControllerDescriptions(const Ice::Current& = GlobalIceCurrent) const override;

            /**
             * @brief Returns the status and description of the \ref NJointController.
             * @param name The \ref NJointController's name.
             * @return The status and description of the \ref NJointController.
             * @see getNJointControllerDescriptionsWithStatuses
             * @see NJointControllerStatus
             * @see getNJointControllerStatus
             * @see getNJointControllerStatuses
             * @see NJointControllerDescription
             * @see getNJointControllerDescription
             * @see getNJointControllerDescriptions
             */
            NJointControllerDescriptionWithStatus getNJointControllerDescriptionWithStatus(
                const std::string& name, const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns the status and description of all \ref NJointController "NJointControllers".
             * @return The status and description of all \ref NJointController "NJointControllers".
             * @see getNJointControllerDescriptionsWithStatus
             * @see NJointControllerStatus
             * @see getNJointControllerStatus
             * @see getNJointControllerStatuses
             * @see NJointControllerDescription
             * @see getNJointControllerDescription
             * @see getNJointControllerDescriptions
             */
            NJointControllerDescriptionWithStatusSeq getNJointControllerDescriptionsWithStatuses(const Ice::Current& = GlobalIceCurrent) const override;

            /**
             * @brief getNJointControllerClassDescription
             * @param className
             * @return
             */
            NJointControllerClassDescription getNJointControllerClassDescription(
                const std::string& className, const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief getNJointControllerClassDescriptions
             * @return
             */
            NJointControllerClassDescriptionSeq getNJointControllerClassDescriptions(const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Loads the given lib. (calls `getArmarXManager()->loadLibFromPath(path)`)
             * @param path Path to the lib to load.
             * @return Whether loading the lib was successful.
             * @see ArmarXManager::loadLibFromPath
             */
            bool loadLibFromPath(const std::string& path, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Loads the given lib. (calls `getArmarXManager()->loadLibFromPath(package, lib)`)
             * @param package The armarx package containing the lib
             * @param lib The lib name to load.
             * @return Whether loading the lib was successful.
             * @see ArmarXManager::loadLibFromPackage
             */
            bool loadLibFromPackage(const std::string& package, const std::string& lib, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Returns the names of all available classes of \ref NJointController.
             * @return The names of all available classes of \ref NJointController.
             */
            Ice::StringSeq getNJointControllerClassNames(const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns the names of all \ref NJointController "NJointControllers"
             * @return The names of all \ref NJointController "NJointControllers"
             */
            Ice::StringSeq getNJointControllerNames(const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns the names of all requested \ref NJointController "NJointControllers"
             * @return The names of all requested \ref NJointController "NJointControllers"
             */
            Ice::StringSeq getRequestedNJointControllerNames(const Ice::Current& = GlobalIceCurrent) const override;
            /**
             * @brief Returns the names of all activated \ref NJointController "NJointControllers"
             * @return The names of all activated \ref NJointController "NJointControllers"
             */
            Ice::StringSeq getActivatedNJointControllerNames(const Ice::Current& = GlobalIceCurrent) const override;

            /**
             * @brief Queues the given \ref NJointController for deletion.
             * @param name The \ref NJointController to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointControllers
             */
            void deleteNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Queues the given \ref NJointController "NJointControllers" for deletion.
             * @param names The \ref NJointController "NJointControllers" to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointController
             */
            void deleteNJointControllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Queues the given \ref NJointController for deletion and deactivates it if necessary.
             * @param name The \ref NJointController to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointController
             * @see deleteNJointControllers
             * @see deactivateAnddeleteNJointControllers
             */
            void deactivateAndDeleteNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Queues the given \ref NJointController "NJointControllers" for deletion and deactivates them if necessary.
             * @param names The \ref NJointController "NJointControllers" to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointController
             * @see deleteNJointControllers
             * @see deactivateAnddeleteNJointController
             */
            void deactivateAndDeleteNJointControllers(const Ice::StringSeq& names, const Ice::Current&) override;

            /**
             * @brief Requests activation for the given \ref NJointController.
             * @param name The requested \ref NJointController.
             * @see activateNJointControllers
             */
            void activateNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Requests activation for the given \ref NJointController "NJointControllers".
             * @param names The requested \ref NJointController "NJointControllers".
             * @see activateNJointController
             */
            void activateNJointControllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;

            /**
             * @brief Requests deactivation for the given \ref NJointController.
             * @param name The \ref NJointController to be deactivated.
             * @see deactivateNJointControllers
             */
            void deactivateNJointController(const std::string& name, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Requests deactivation for the given \ref NJointController "NJointControllers".
             * @param names The \ref NJointController "NJointControllers" to be deactivated.
             * @see deactivateNJointController
             */
            void deactivateNJointControllers(const Ice::StringSeq& names, const Ice::Current& = GlobalIceCurrent) override;

            /**
             * @brief Cretes a \ref NJointController.
             * @param className The \ref NJointController's class.
             * @param instanceName The \ref NJointController's name.
             * @param config A config passed to the \ref NJointController's ctor.
             * @return A proxy to the created \ref NJointController.
             */
            NJointControllerInterfacePrx createNJointController(
                const std::string& className, const std::string& instanceName,
                const NJointControllerConfigPtr& config, const Ice::Current& = GlobalIceCurrent) override;
            /**
             * @brief Cretes a \ref NJointController.
             * @param className The \ref NJointController's class.
             * @param instanceName The \ref NJointController's name.
             * @param variants A map of \ref Variant "Variants" passed to the \ref NJointController's ctor.
             * @return A proxy to the created \ref NJointController.
             */
            NJointControllerInterfacePrx createNJointControllerFromVariantConfig(
                const std::string& className, const std::string& instanceName,
                const StringVariantBaseMap& variants, const Ice::Current& = GlobalIceCurrent) override;

            /**
             * @brief Changes the set of requested \ref NJointController "NJointControllers" to the given set.
             * @param newSetup The new set of requested \ref NJointController "NJointControllers"
             */
            void switchNJointControllerSetup(const Ice::StringSeq& newSetup, const Ice::Current& = GlobalIceCurrent) override;
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // /////////////////////////////////// Module interface /////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        public:
            /**
             * @brief Cretes a \ref NJointController.
             * @param className The \ref NJointController's class.
             * @param instanceName The \ref NJointController's name.
             * @param config A config passed to the \ref NJointController's ctor.
             * @param deletable Whether the \ref NJointController can be deleted.
             * @param internal Whether the \ref NJointController should be tagged as internal.
             * @return A pointer to the created \ref NJointController
             */
            const NJointControllerPtr& createNJointController(
                const std::string& className, const std::string& instanceName,
                const NJointControllerConfigPtr& config, bool deletable, bool internal);

            /**
             * @brief Returns pointers to the \ref NJointController "NJointControllers". (If one does not exist, an exception is thrown.)
             * @param names The \ref NJointController "NJointControllers" names.
             * @return Pointers to the \ref NJointController "NJointControllers".
             * @throw If there is no \ref NJointcontroller for \param name
             */
            std::vector<armarx::NJointControllerPtr> getNJointControllersNotNull(const std::vector<std::string>& names) const;
            /**
             * @brief Returns a pointer to the \ref NJointController. (If it does not exist, an exception is thrown.)
             * @param name The \ref NJointController
             * @return A pointer to the \ref NJointController.
             * @throw If there is no \ref NJointcontroller for \param name
             */
            const NJointControllerPtr& getNJointControllerNotNull(const std::string& name) const;

            /**
             * @brief Returns the names of given \ref NJointController "NJointControllers"
             * @param ctrls The \ref NJointController "NJointControllers"
             * @return The names of given \ref NJointController "NJointControllers"
             */
            std::vector<std::string> getNJointControllerNames(const std::vector<armarx::NJointControllerPtr>& ctrls) const;

            /**
             * @brief Queues the given \ref NJointController for deletion.
             * @param ctrl The \ref NJointController to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointControllers
             */
            void deleteNJointController(const NJointControllerPtr& ctrl);
            /**
             * @brief Queues the given \ref NJointController "NJointControllers" for deletion.
             * @param ctrls The \ref NJointController "NJointControllers" to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointController
             */
            void deleteNJointControllers(const std::vector<NJointControllerPtr>& ctrls);
            /**
             * @brief Queues the given \ref NJointController for deletion and deactivates it if necessary.
             * @param ctrl The \ref NJointController to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointController
             * @see deleteNJointControllers
             * @see deactivateAnddeleteNJointControllers
             */
            void deactivateAndDeleteNJointController(const NJointControllerPtr& ctrl);
            /**
             * @brief Queues the given \ref NJointController "NJointControllers" for deletion and deactivates them if necessary.
             * @param ctrls The \ref NJointController "NJointControllers" to delete.
             * @see removeNJointControllersToBeDeleted
             * @see nJointControllersToBeDeleted
             * @see deleteNJointController
             * @see deleteNJointControllers
             * @see deactivateAnddeleteNJointController
             */
            void deactivateAndDeleteNJointControllers(const std::vector<NJointControllerPtr>& ctrls);


            /**
             * @brief Requests activation for the given \ref NJointController.
             * @param ctrl The requested \ref NJointController.
             * @see activateNJointControllers
             */
            void activateNJointController(const NJointControllerPtr& ctrl);
            /**
             * @brief Requests activation for the given \ref NJointController "NJointControllers".
             * @param ctrls The requested \ref NJointController "NJointControllers".
             * @see activateNJointController
             */
            void activateNJointControllers(const std::vector<NJointControllerPtr>& ctrls);

            /**
             * @brief Requests deactivation for the given \ref NJointController.
             * @param ctrl The \ref NJointController to be deactivated.
             * @see deactivateNJointControllers
             */
            void deactivateNJointController(const NJointControllerPtr& ctrl);
            /**
             * @brief Requests deactivation for the given \ref NJointController "NJointControllers".
             * @param ctrls The \ref NJointController "NJointControllers" to be deactivated.
             * @see deactivateNJointController
             */
            void deactivateNJointControllers(const std::vector<NJointControllerPtr>& ctrls);

            // //////////////////////////////////////////////////////////////////////////////////////// //
            // //////////////////////////////////// implementation //////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            /**
             * @brief Deletes the given \ref NJointController "NJointControllers" and removes them
             * from the \ref ArmarXManager
             * @param ctrls \ref NJointController "NJointControllers" to delete.
             * @param blocking Whether removal from the \ref ArmarXManager should be blocking.
             * @param l Proxy to the listener topic all removal events should be sent to.
             */
            void removeNJointControllers(std::map<std::string, NJointControllerPtr>& ctrls, bool blocking = true, RobotUnitListenerPrx l = nullptr);
            /**
             * @brief Calls \ref removeNJointControllers for all \ref NJointController "NJointControllers"
             * queued for deletion (\ref nJointControllersToBeDeleted).
             * @param blocking Whether removal from the \ref ArmarXManager should be blocking.
             * @param l Proxy to the listener topic all removal events should be sent to.
             * @see removeNJointControllers
             * @see nJointControllersToBeDeleted
             */
            void removeNJointControllersToBeDeleted(bool blocking = true, RobotUnitListenerPrx l = nullptr);

            /**
             * @brief Sets the requested flag for all given \ref NJointController "NJointControllers" and unsets it for the rest.
             * @param request The \ref NJointController "NJointControllers" where the requested flag should be set.
             */
            void updateNJointControllerRequestedState(const std::set<NJointControllerPtr>& request);
            /**
             * @brief Throws if there is no factory for \ref NJointController "NJointControllers" for the given class name.
             * @param className The class to check for.
             * @throw If there is no factory for \ref NJointController "NJointControllers" for the given class name.
             */
            void checkNJointControllerClassName(const std::string& className) const;

            // //////////////////////////////////////////////////////////////////////////////////////// //
            // ///////////////////////////////////////// Data ///////////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            /// @brief Holds all currently loaded NJoint controllers (index: [instancename])(May not be accessed in rt.)
            std::map<std::string, NJointControllerPtr> nJointControllers;
            /// @brief These controllers will be deleted in the next iteration of publish
            std::map<std::string, NJointControllerPtr> nJointControllersToBeDeleted;
            /// @brief VirtualRobot used when creating controllers / calling other functions in this module
            VirtualRobot::RobotPtr controllerCreateRobot;

            std::recursive_mutex controllerMutex;
            // //////////////////////////////////////////////////////////////////////////////////////// //
            // /////////////////////////////////////// Attorneys ////////////////////////////////////// //
            // //////////////////////////////////////////////////////////////////////////////////////// //
        private:
            /**
            * \brief This class allows minimal access to private members of \ref ControllerManagement in a sane fashion for \ref Publisher.
            * \warning !! DO NOT ADD ADDITIONAL FRIENDS IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
            */
            friend class ControllerManagementAttorneyForPublisher;
            /**
            * \brief This class allows minimal access to private members of \ref ControllerManagement in a sane fashion for \ref ControlThreadDataBuffer.
            * \warning !! DO NOT ADD ADDITIONAL FRIENDS IF YOU DO NOT KNOW WAHT YOU ARE DOING! IF YOU DO SOMETHING WRONG YOU WILL CAUSE UNDEFINED BEHAVIOUR !!
            */
            friend class ControllerManagementAttorneyForControlThreadDataBuffer;
        };
    }
}
