/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::RobotUnit
 * @author     Raphael Grimm ( raphael dot grimm at kit dot edu )
 * @date       2018
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <ArmarXCore/core/util/FileSystemPathBuilder.h>
#include <ArmarXCore/core/ArmarXManager.h>

#include "RobotUnitModuleLogging.h"

#include "RobotUnitModuleControlThreadDataBuffer.h"
#include "RobotUnitModuleDevices.h"

#include "../util/ControlThreadOutputBuffer.h"

namespace armarx
{
    namespace RobotUnitModule
    {
        void Logging::addMarkerToRtLog(const RemoteReferenceCounterBasePtr& token, const std::string& marker, const Ice::Current&)
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            std::lock_guard<std::mutex> guard {rtLoggingMutex};
            if (!rtLoggingEntries.count(token->getId()))
            {
                throw InvalidArgumentException {"addMarkerToRtLog called for a nonexistent log"};
            }
            rtLoggingEntries.at(token->getId())->marker = marker;
        }

        RemoteReferenceCounterBasePtr Logging::startRtLogging(const std::string& formatString, const Ice::StringSeq& loggingNames, const Ice::Current&)
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            StringStringDictionary alias;
            for (const auto& name : loggingNames)
            {
                alias.emplace(name, "");
            }
            return startRtLoggingWithAliasNames(formatString, alias, GlobalIceCurrent);
        }

        void Logging::stopRtLogging(const RemoteReferenceCounterBasePtr& token, const Ice::Current&)
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            std::lock_guard<std::mutex> guard {rtLoggingMutex};
            if (!rtLoggingEntries.count(token->getId()))
            {
                throw InvalidArgumentException {"addMarkerToRtLog called for a nonexistent log"};
            }
            ARMARX_DEBUG_S << "RobotUnit: stop RtLogging for file " << rtLoggingEntries.at(token->getId())->filename;
            rtLoggingEntries.at(token->getId())->stopLogging = true;
        }

        Ice::StringSeq Logging::getLoggingNames(const Ice::Current&) const
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            Ice::StringSeq result;
            for (const auto& strs : sensorDeviceValueLoggingNames)
            {
                result.insert(result.end(), strs.begin(), strs.end());
            }
            for (const auto& strvecss : controlDeviceValueLoggingNames)
            {
                for (const auto& strs : strvecss)
                {
                    result.insert(result.end(), strs.begin(), strs.end());
                }
            }
            return result;
        }

        RemoteReferenceCounterBasePtr Logging::startRtLoggingWithAliasNames(const std::string& formatString, const StringStringDictionary& aliasNames, const Ice::Current&)
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            FileSystemPathBuilder pb {formatString};
            std::lock_guard<std::mutex> guard {rtLoggingMutex};
            if (rtLoggingEntries.count(pb.getPath()))
            {
                throw InvalidArgumentException {"There already is a logger for the path '" + pb.getPath() + "'"};
            }
            auto ptr = rtLoggingEntries[pb.getPath()];
            CSVLoggingEntry& e = *ptr;
            e.filename = pb.getPath();
            pb.createParentDirectories();
            e.stream.open(e.filename);
            if (!e.stream)
            {
                rtLoggingEntries.erase(pb.getPath());
                throw LogicError {"RtLogging could not open filestream for '" + pb.getPath() + "'"};
            }

            std::stringstream header;
            header << "marker;iteration;timestamp";
            auto logDev = [&](const std::string & dev)
            {
                for (const auto& pair : aliasNames)
                {
                    std::string name = pair.first;
                    if (boost::starts_with(dev + "$", name))
                    {
                        if ((name.size() > 0 && name.back() == '$'))
                        {
                            name.resize(name.size() - 1);
                        }
                        const std::string& alias = pair.second.empty() ? name : pair.second;
                        header << ";" << alias << dev.substr(name.size());
                        return true;
                    }
                }
                return false;
            };

            //get logged sensor device values
            {
                e.loggedSensorDeviceValues.reserve(sensorDeviceValueLoggingNames.size());
                for (const auto& svfields : sensorDeviceValueLoggingNames)
                {
                    e.loggedSensorDeviceValues.emplace_back();
                    auto& svfieldsFlags = e.loggedSensorDeviceValues.back(); //vv
                    svfieldsFlags.reserve(svfields.size());
                    for (const auto& field : svfields)
                    {
                        svfieldsFlags.emplace_back(logDev(field));
                    }
                }
            }
            //get logged control device values
            {
                e.loggedControlDeviceValues.reserve(controlDeviceValueLoggingNames.size());
                for (const auto& sjctrls : controlDeviceValueLoggingNames)
                {
                    e.loggedControlDeviceValues.emplace_back();
                    auto& sjctrlsFlags = e.loggedControlDeviceValues.back(); //vv
                    sjctrlsFlags.reserve(sjctrls.size());
                    for (const auto& ctargfields : sjctrls)
                    {

                        sjctrlsFlags.emplace_back();
                        auto& ctargfieldsFlags = sjctrlsFlags.back(); //v
                        ctargfieldsFlags.reserve(ctargfields.size());

                        for (const auto& field : ctargfields)
                        {
                            ctargfieldsFlags.emplace_back(logDev(field));
                        }
                    }
                }
            }

            //write header
            e.stream << header.str() << std::flush; // newline is written at the beginning of each log line
            //create and return handle
            auto block = getArmarXManager()->createRemoteReferenceCount(
                             [ptr]()
            {
                ARMARX_DEBUG_S << "RobotUnit: stop RtLogging for file " << ptr->filename;
                ptr->stopLogging = true;
            }, e.filename);
            auto counter = block->getReferenceCounter();
            block->activateCounting();
            ARMARX_DEBUG_S << "RobotUnit: start RtLogging for file " << ptr->filename;
            return counter;
        }

        void Logging::writeRecentIterationsToFile(const std::string& formatString, const Ice::Current&) const
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            std::lock_guard<std::mutex> guard {rtLoggingMutex};
            FileSystemPathBuilder pb {formatString};
            pb.createParentDirectories();
            std::ofstream outCSV {pb.getPath() + ".csv"};
            if (!outCSV)
            {
                throw LogicError {"writeRecentIterationsToFile could not open filestream for '" + pb.getPath() + ".csv'"};
            }
            std::ofstream outMsg {pb.getPath() + ".messages"};
            if (!outMsg)
            {
                throw LogicError {"writeRecentIterationsToFile could not open filestream for '" + pb.getPath() + ".messages'"};
            }
            ARMARX_INFO << "writing the last " << backlog.size() << " iterations to " << pb.getPath() << ".{csv, messages}";
            //write csv header
            {
                outCSV << "iteration;timestamp";
                for (const auto& vs : sensorDeviceValueLoggingNames)
                {
                    for (const auto& s : vs)
                    {
                        outCSV << ";" << s;
                    }
                }
                for (const auto& vvs : controlDeviceValueLoggingNames)
                {
                    for (const auto& vs : vvs)
                    {
                        for (const auto& s : vs)
                        {
                            outCSV << ";" << s;
                        }
                    }
                }
                outCSV << std::endl;
            }

            for (const ::armarx::detail::ControlThreadOutputBufferEntry& iteration : backlog)
            {
                //write csv data
                {
                    outCSV  << iteration.iteration << ";" << iteration.sensorValuesTimestamp;
                    //sens
                    {
                        for (const SensorValueBase* val : iteration.sensors)
                        {
                            for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                            {
                                outCSV  << ";" << val->getDataFieldAsString(idxField);
                            }
                        }
                    }
                    //ctrl
                    {
                        for (const auto& vals : iteration.control)
                        {
                            for (const ControlTargetBase* val : vals)
                            {
                                for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                                {
                                    outCSV  << ";" << val->getDataFieldAsString(idxField);
                                }
                            }
                        }
                    }
                    outCSV << std::endl;
                }
                //write message data
                {
                    for (const ::armarx::detail::RtMessageLogEntryBase* msg : iteration.messages.getEntries())
                    {
                        if (!msg)
                        {
                            break;
                        }
                        outMsg << "[" << msg->getTime().toDateTime() << "] iteration "
                               << iteration.iteration << ":\n"
                               << msg->format() << std::endl;
                    }
                    outMsg << "\nmessages lost: " << iteration.messages.messagesLost
                           << " (required additional "
                           << iteration.messages.requiredAdditionalBufferSpace << " bytes, "
                           << iteration.messages.requiredAdditionalEntries << " message entries)\n" << std::endl;
                }
            }
        }

        void Logging::_preFinishRunning()
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            defaultLogHandle = nullptr;
            if (rtLoggingTask)
            {
                ARMARX_DEBUG << "shutting down rt logging task";
                rtLoggingTask->stop();
                while (rtLoggingTask->isFunctionExecuting() || rtLoggingTask->isRunning())
                {
                    ARMARX_FATAL << deactivateSpam(0.1) << "RT LOGGING TASK IS RUNNING AFTER IT WAS STOPPED!";
                }
                rtLoggingTask = nullptr;
                ARMARX_DEBUG << "shutting down rt logging task done";
            }
        }

        void Logging::_preFinishControlThreadInitialization()
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            controlThreadId = LogSender::getThreadId();
            ControlThreadOutputBuffer::RtLoggingInstance = &(_module<ControlThreadDataBuffer>().getControlThreadOutputBuffer());

            ARMARX_INFO << "starting rt logging with timestep " << rtLoggingTimestep;
            rtLoggingTask->start();
        }

        void Logging::doLogging()
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            std::lock_guard<std::mutex> guard {rtLoggingMutex};
            const auto now = IceUtil::Time::now();
            // entries are removed last

            //remove backlog entries
            {
                while (!backlog.empty() && backlog.front().writeTimestamp + rtLoggingBacklogRetentionTime < now)
                {
                    backlog.pop_front();
                }
            }
            //log all
            {
                _module<ControlThreadDataBuffer>().getControlThreadOutputBuffer().foreachNewLoggingEntry(
                    [&](const ControlThreadOutputBuffer::Entry & data)
                {
                    //base (marker;iteration;timestamp)
                    {
                        for (auto& pair : rtLoggingEntries)
                        {
                            CSVLoggingEntry& e = *pair.second;
                            e.stream << "\n"
                                     << e.marker << ";"
                                     << data.iteration << ";"
                                     << data.sensorValuesTimestamp.toMicroSeconds();
                            e.marker.clear();
                        }
                    }
                    //sens
                    {
                        for (std::size_t idxDev = 0; idxDev < data.sensors.size(); ++ idxDev)
                        {
                            const SensorValueBase* val = data.sensors.at(idxDev);
                            for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                            {
                                const auto str = val->getDataFieldAsString(idxField);
                                for (auto& pair : rtLoggingEntries)
                                {
                                    CSVLoggingEntry& e = *pair.second;
                                    if (e.loggedSensorDeviceValues.at(idxDev).at(idxField))
                                    {
                                        pair.second->stream  << ";" << str;
                                    }
                                }
                            }
                        }
                    }
                    //ctrl
                    {
                        for (std::size_t idxDev = 0; idxDev < data.control.size(); ++ idxDev)
                        {
                            const auto& vals = data.control.at(idxDev);
                            for (std::size_t idxVal = 0; idxVal < vals.size(); ++idxVal)
                            {
                                const ControlTargetBase* val = vals.at(idxVal);
                                for (std::size_t idxField = 0; idxField < val->getNumberOfDataFields(); ++ idxField)
                                {
                                    const auto str = val->getDataFieldAsString(idxField);
                                    for (auto& pair : rtLoggingEntries)
                                    {
                                        CSVLoggingEntry& e = *pair.second;
                                        if (e.loggedControlDeviceValues.at(idxDev).at(idxVal).at(idxField))
                                        {
                                            pair.second->stream  << ";" << str;
                                        }
                                    }
                                }
                            }
                        }
                    }
                    //store data to backlog
                    {
                        if (data.writeTimestamp + rtLoggingBacklogRetentionTime >= now)
                        {
                            backlog.emplace_back(data, true); //true for minimal copy
                        }
                    }
                    //print + reset messages
                    {
                        for (const ::armarx::detail::RtMessageLogEntryBase* ptr : data.messages.getEntries())
                        {
                            if (!ptr)
                            {
                                break;
                            }
                            ptr->print(controlThreadId);
                        }
                    }
                });
            }
            ARMARX_DEBUG_S << ::deactivateSpam() << "the last " << backlog.size() << " iterations are stored";
            //flush all files
            {
                for (auto& pair : rtLoggingEntries)
                {
                    pair.second->stream << std::flush;
                }
            }

            //remove entries
            {
                std::vector<std::string> toRemove;
                toRemove.reserve(rtLoggingEntries.size());
                for (auto& pair : rtLoggingEntries)
                {
                    if (pair.second->stopLogging)
                    {
                        //can't remove the current elemet
                        //(this would invalidate the current iterator)
                        toRemove.emplace_back(pair.first);
                    }
                }
                for (const auto& rem : toRemove)
                {
                    rtLoggingEntries.erase(rem);
                }
            }
        }

        void Logging::_postOnInitRobotUnit()
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            rtLoggingTimestep = getProperty<std::size_t>("RTLogging_PeriodMs");
            ARMARX_CHECK_LESS_W_HINT(0, rtLoggingTimestep, "The property RTLoggingPeriodMs must not be 0");

            messageBufferSize = getProperty<std::size_t>("RTLogging_MessageBufferSize");
            messageBufferMaxSize = getProperty<std::size_t>("RTLogging_MaxMessageBufferSize");
            ARMARX_CHECK_LESS_EQUAL(messageBufferSize, messageBufferMaxSize);

            messageBufferNumberEntries = getProperty<std::size_t>("RTLogging_MessageNumber");
            messageBufferMaxNumberEntries = getProperty<std::size_t>("RTLogging_MaxMessageNumber");
            ARMARX_CHECK_LESS_EQUAL(messageBufferNumberEntries, messageBufferMaxNumberEntries);

            rtLoggingBacklogRetentionTime = IceUtil::Time::milliSeconds(getProperty<std::size_t>("RTLogging_KeepIterationsForMs"));

            ARMARX_CHECK_GREATER(getControlThreadTargetPeriod().toMicroSeconds(), 0);
        }

        void Logging::_postFinishDeviceInitialization()
        {
            throwIfInControlThread(BOOST_CURRENT_FUNCTION);
            //init buffer
            {
                std::size_t ctrlThreadPeriodUs = static_cast<std::size_t>(getControlThreadTargetPeriod().toMicroSeconds());
                std::size_t logThreadPeriodUs = rtLoggingTimestep * 1000;
                std::size_t nBuffers = (logThreadPeriodUs / ctrlThreadPeriodUs + 1) * 100;

                const auto bufferSize = _module<ControlThreadDataBuffer>().getControlThreadOutputBuffer().initialize(
                                            nBuffers, _module<Devices>().getControlDevices(), _module<Devices>().getSensorDevices(),
                                            messageBufferSize, messageBufferNumberEntries,
                                            messageBufferMaxSize, messageBufferMaxNumberEntries);
                ARMARX_INFO << "RTLogging activated! using "
                            << nBuffers << "buffers "
                            << "(buffersize = " << bufferSize << " bytes)";
            }
            //init logging names
            {
                //sensorDevices
                controlDeviceValueLoggingNames.reserve(_module<Devices>().getControlDevices().size());
                for (const auto& cd : _module<Devices>().getControlDevices().values())
                {
                    controlDeviceValueLoggingNames.emplace_back();
                    auto& namesForDev = controlDeviceValueLoggingNames.back();
                    namesForDev.reserve(cd->getJointControllers().size());
                    for (auto jointC : cd->getJointControllers())
                    {
                        const auto names = jointC->getControlTarget()->getDataFieldNames();
                        namesForDev.emplace_back();
                        auto& fullNames = namesForDev.back();
                        fullNames.reserve(names.size());
                        for (const auto& name :  names)
                        {
                            fullNames.emplace_back(
                                "ctrl." +
                                cd->getDeviceName() + "." +
                                jointC->getControlMode() + "." +
                                name);
                        }
                    }
                }
                //sensorDevices
                sensorDeviceValueLoggingNames.reserve(_module<Devices>().getSensorDevices().size());
                for (const auto& sd : _module<Devices>().getSensorDevices().values())
                {
                    const auto names = sd->getSensorValue()->getDataFieldNames();
                    sensorDeviceValueLoggingNames.emplace_back();
                    auto& fullNames = sensorDeviceValueLoggingNames.back();
                    fullNames.reserve(names.size());
                    for (const auto& name :  names)
                    {
                        fullNames.emplace_back(
                            "sens." +
                            sd->getDeviceName() + "." +
                            name);
                    }
                }
            }
            //start logging thread is done in rtinit
            //maybe add the default log
            {
                const auto loggingpath = getProperty<std::string>("RTLogging_DefaultLog").getValue();
                if (!loggingpath.empty())
                {
                    defaultLogHandle = startRtLogging(loggingpath, getLoggingNames());
                }
            }
            //setup thread (don't start it
            rtLoggingTask = new RTLoggingTaskT([&] {doLogging();}, rtLoggingTimestep, false, getName() + "_RTLoggingTask");
            rtLoggingTask->setDelayWarningTolerance(rtLoggingTimestep * 10);
        }
    }
}
