/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::TCPControllerSubUnit
 * @author     Stefan Reither ( stef dot reither at web dot de )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "TCPControllerSubUnit.h"

#include <RobotAPI/libraries/core/FramedPose.h>

#include <RobotAPI/components/units/RobotUnit/NJointControllers/NJointCartesianVelocityController.h>

using namespace armarx;

void TCPControllerSubUnit::update(const SensorAndControl& sc, const JointAndNJointControllers& c)
{

}

PropertyDefinitionsPtr TCPControllerSubUnit::createPropertyDefinitions()
{
    return PropertyDefinitionsPtr(new TCPControllerSubUnitPropertyDefinitions(getConfigIdentifier()));
}

void TCPControllerSubUnit::setup(RobotUnit* rUnit, VirtualRobot::RobotPtr robot)
{
    ARMARX_CHECK_EXPRESSION(robot);
    ARMARX_CHECK_EXPRESSION(!this->coordinateTransformationRobot);
    coordinateTransformationRobot = robot;

    ARMARX_CHECK_EXPRESSION(!robotUnit);
    ARMARX_CHECK_EXPRESSION(rUnit);
    robotUnit = rUnit;
}

void TCPControllerSubUnit::setCycleTime(Ice::Int, const Ice::Current& c)
{
    ARMARX_WARNING << "Setting cycle time in RT-Controller does not have an effect";
}

void TCPControllerSubUnit::setTCPVelocity(const std::string& nodeSetName, const std::string& tcpName, const FramedDirectionBasePtr& translationVelocity, const FramedDirectionBasePtr& orientationVelocityRPY, const Ice::Current& c)
{
    ARMARX_CHECK_EXPRESSION_W_HINT(coordinateTransformationRobot->hasRobotNodeSet(nodeSetName), "The robot does not have the node set: " + nodeSetName);
    std::string tcp;
    if (tcpName.empty())
    {
        tcp = coordinateTransformationRobot->getRobotNodeSet(nodeSetName)->getTCP()->getName();
    }
    else
    {
        ARMARX_CHECK_EXPRESSION_W_HINT(coordinateTransformationRobot->hasRobotNode(tcpName), "The robot does not have the node: " + tcpName);
        tcp = tcpName;
    }

    auto activeNJointControllers = robotUnit->getNJointControllersNotNull(robotUnit->getNJointControllerNames());
    NJointCartesianVelocityControllerPtr tcpController;
    bool nodeSetAlreadyControlled = false;
    for (NJointControllerPtr controller : activeNJointControllers)
    {
        tcpController = NJointCartesianVelocityControllerPtr::dynamicCast(controller);
        if (!tcpController)
        {
            continue;
        }
        if (tcpController->getNodeSetName() == nodeSetName)
        {
            nodeSetAlreadyControlled = true;
            break;
        }
    }
    robotUnit->updateVirtualRobot(coordinateTransformationRobot);

    float xVel = 0.f;
    float yVel = 0.f;
    float zVel = 0.f;
    float rollVel = 0.f;
    float pitchVel = 0.f;
    float yawVel = 0.f;


    FramedDirectionPtr globalTransVel = FramedDirectionPtr::dynamicCast(translationVelocity);
    if (globalTransVel)
    {
        globalTransVel->changeFrame(coordinateTransformationRobot, coordinateTransformationRobot->getRootNode()->getName());
        xVel = globalTransVel->x;
        yVel = globalTransVel->y;
        zVel = globalTransVel->z;
    }

    FramedDirectionPtr globalOriVel = FramedDirectionPtr::dynamicCast(orientationVelocityRPY);
    if (globalOriVel)
    {
        globalOriVel->changeFrame(coordinateTransformationRobot, coordinateTransformationRobot->getRootNode()->getName());
        rollVel = globalOriVel->x;
        pitchVel = globalOriVel->y;
        yawVel = globalOriVel->z;

    }

    NJointCartesianVelocityControllerMode::CartesianSelection mode = NJointCartesianVelocityControllerMode::eAll;
    bool noMode = false;
    if (globalTransVel && globalOriVel)
    {
        mode = NJointCartesianVelocityControllerMode::eAll;
    }
    else if (globalTransVel && !globalOriVel)
    {
        mode = NJointCartesianVelocityControllerMode::ePosition;
    }
    else if (!globalTransVel && globalOriVel)
    {
        mode = NJointCartesianVelocityControllerMode::eOrientation;
    }
    else
    {
        noMode = true;
    }
    ARMARX_DEBUG << "CartesianSelection-Mode: " << (int)mode;

    auto controllerName = this->getName() + "_" + tcp + "_" + nodeSetName;
    if (!nodeSetAlreadyControlled)
    {
        NJointCartesianVelocityControllerConfigPtr config = new NJointCartesianVelocityControllerConfig(nodeSetName, tcp, mode);
        NJointCartesianVelocityControllerPtr ctrl = NJointCartesianVelocityControllerPtr::dynamicCast(
                    robotUnit->createNJointController(
                        "NJointCartesianVelocityController", controllerName,
                        config, true, true
                    )
                );

        tcpController = ctrl;
        tcpController->setAvoidJointLimitsKp(getProperty<float>("AvoidJointLimitsKp").getValue());
    }



    // Only activate controller if at least either translationVelocity or orientaionVelocity is set
    if (!noMode)
    {
        ARMARX_CHECK_EXPRESSION(tcpController);
        //        ARMARX_CHECK_EXPRESSION(tcpController->getObjectScheduler());
        tcpController->setVelocities(xVel, yVel, zVel, rollVel, pitchVel, yawVel, NJointCartesianVelocityController::ModeFromIce(mode));
        //        if (!tcpController->getObjectScheduler()->waitForObjectStateMinimum(eManagedIceObjectStarted, 5000))
        //        {
        //            ARMARX_ERROR << "NJointController was not initialized after 5000ms - bailing out!";
        //            return;
        //        }
        if (!tcpController->isControllerActive())
        {
            robotUnit->activateNJointController(controllerName);
        }
    }
}

bool TCPControllerSubUnit::isRequested(const Ice::Current&)
{
    auto activeNJointControllers = robotUnit->getNJointControllersNotNull(robotUnit->getNJointControllerNames());
    for (NJointControllerPtr controller : activeNJointControllers)
    {
        auto tcpController = NJointCartesianVelocityControllerPtr::dynamicCast(controller);
        if (!tcpController)
        {
            return true;
        }
    }

    return false;
}

void armarx::TCPControllerSubUnit::componentPropertiesUpdated(const std::set<std::string>& changedProperties)
{
    if (changedProperties.count("AvoidJointLimitsKp") && robotUnit)
    {
        float avoidJointLimitsKp = getProperty<float>("AvoidJointLimitsKp").getValue();
        auto activeNJointControllers = robotUnit->getNJointControllersNotNull(robotUnit->getNJointControllerNames());
        for (NJointControllerPtr controller : activeNJointControllers)
        {
            auto tcpController = NJointCartesianVelocityControllerPtr::dynamicCast(controller);
            if (tcpController)
            {
                tcpController->setAvoidJointLimitsKp(avoidJointLimitsKp);
            }
        }
    }
}
