/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::GamepadUnit
 * @author     Simon Ottenhaus ( simon dot ottenhaus at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "GamepadUnit.h"

#include <ArmarXCore/observers/variant/TimestampVariant.h>
#include <linux/joystick.h>

using namespace armarx;

void GamepadUnit::onInitComponent()
{
    offeringTopic(getProperty<std::string>("GamepadTopicName").getValue());
    deviceName = getProperty<std::string>("GamepadDeviceName").getValue();
    readTask = new RunningTask<GamepadUnit>(this, &GamepadUnit::run, "GamepadUnit");
}

void GamepadUnit::onConnectComponent()
{
    topicPrx = getTopic<GamepadUnitListenerPrx>(getProperty<std::string>("GamepadTopicName").getValue());
    if (js.open(deviceName))
    {
        ARMARX_INFO << "opened a gamepad named " << js.name << " with " << js.numberOfAxis << " axis and " << js.numberOfButtons << " buttons.";
        if (js.numberOfAxis == 8 && js.numberOfButtons == 11)
        {
            readTask->start();
        }
        else
        {
            ARMARX_WARNING << "this is not our trusty logitech gamepad you moron.";
        }
    }
    else
    {
        ARMARX_WARNING << "Could not open gamepad device " << deviceName;
    }
}

void GamepadUnit::run()
{
    while (readTask->isRunning())
    {
        GamepadData data;

        if (!js.pollEvent())
        {
            ARMARX_WARNING << "failed to read gamepad data";
            break;
        }

        IceUtil::Time now = IceUtil::Time::now();
        TimestampVariantPtr nowTimestamp = new TimestampVariant(now);

        //mapping found with command line tool jstest <device file>

        float axisFactor = 1.0f / 32768.f;

        data.leftStickY = js.axis[0] * axisFactor;
        data.leftStickX = js.axis[1] * axisFactor;
        data.rightStickX = js.axis[3] * axisFactor;
        data.rightStickY = js.axis[4] * axisFactor;
        data.dPadX = js.axis[7] * axisFactor;
        data.dPadY = js.axis[6] * axisFactor;
        data.leftTrigger = js.axis[2] * axisFactor;
        data.rightTrigger = js.axis[5] * axisFactor;

        data.leftButton = js.buttonsPressed[4];
        data.rightButton = js.buttonsPressed[5];
        data.backButton = js.buttonsPressed[6];
        data.startButton = js.buttonsPressed[7];
        data.xButton = js.buttonsPressed[2];
        data.yButton = js.buttonsPressed[3];
        data.aButton = js.buttonsPressed[0];
        data.bButton = js.buttonsPressed[1];
        data.theMiddleButton = js.buttonsPressed[8];
        data.leftStickButton = js.buttonsPressed[9];
        data.rightStickButton = js.buttonsPressed[10];

        ARMARX_IMPORTANT << "left x (integer): " << js.axis[0] << " left x (float): " << data.leftStickX;

        topicPrx->reportGamepadState(deviceName, js.name, data, nowTimestamp);

        //usleep(1000); // 10ms
    }
}


void GamepadUnit::onDisconnectComponent()
{

}


void GamepadUnit::onExitComponent()
{

}

armarx::PropertyDefinitionsPtr GamepadUnit::createPropertyDefinitions()
{
    return armarx::PropertyDefinitionsPtr(new GamepadUnitPropertyDefinitions(
            getConfigIdentifier()));
}

