/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::ArmarXObjects::HokuyoLaserUnit
 * @author     Fabian Paus ( fabian dot paus at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "HokuyoLaserUnit.h"

#include <ArmarXCore/observers/variant/TimestampVariant.h>

#include <boost/algorithm/string/split.hpp>
#include <HokuyoLaserScannerDriver/urg_utils.h>

using namespace armarx;


void HokuyoLaserUnit::onInitComponent()
{
    topicName = getProperty<std::string>("LaserScannerTopicName").getValue();
    offeringTopic(topicName);
    ARMARX_INFO << "Going to report on topic " << topicName;
    updatePeriod = getProperty<int>("UpdatePeriod").getValue();
    angleOffset = getProperty<float>("AngleOffset").getValue();

    std::string deviceStrings = getProperty<std::string>("Devices").getValue();
    std::vector<std::string> splitDeviceStrings;
    boost::split(splitDeviceStrings, deviceStrings, boost::is_any_of(";"));
    devices.clear();
    devices.reserve(splitDeviceStrings.size());
    for (std::string const& deviceString : splitDeviceStrings)
    {
        std::vector<std::string> deviceInfo;
        boost::split(deviceInfo, deviceString, boost::is_any_of(","));
        if (deviceInfo.size() != 3)
        {
            ARMARX_WARNING << "Unexpected format for laser scanner device: " << deviceString
                           << " (split size: " << deviceInfo.size() << ")";
            continue;
        }

        try
        {
            HokuyoLaserScanDevice device;
            device.ip = deviceInfo[0];
            device.port = std::stoi(deviceInfo[1]);
            device.frame = deviceInfo[2];
            device.connected = false;
            devices.push_back(device);
        }
        catch (std::exception const& ex)
        {
            ARMARX_WARNING << "Could not convert port to integer for laser scanner device " << deviceString
                           << " (port is " << deviceInfo[1] << ") : " << ex.what();
            continue;
        }
    }
}


void HokuyoLaserUnit::onConnectComponent()
{
    topic = getTopic<LaserScannerUnitListenerPrx>(topicName);

    if (task)
    {
        task->stop();
    }

    connectedDevices.clear();
    for (HokuyoLaserScanDevice& device : devices)
    {
        ARMARX_INFO << "Connecting to " << device.ip << ":" << device.port;
        int ret = urg_open(&device.urg, URG_ETHERNET, device.ip.c_str(), device.port);
        device.connected = (ret == 0);
        if (!device.connected)
        {
            ARMARX_WARNING << "Could not connect to laser scanner device using URG driver (IP: "
                           << device.ip << ", Port: " << device.port << ", Error: " << ret << ")";
            continue;
        }

        LaserScannerInfo info;
        info.device = device.ip;
        info.frame = device.frame;
        int minStep = 0, maxStep = 0;
        urg_step_min_max(&device.urg, &minStep, &maxStep);
        info.minAngle = urg_step2rad(&device.urg, minStep);
        info.maxAngle = urg_step2rad(&device.urg, maxStep);


        int lengthDataSize = urg_max_data_size(&device.urg);
        info.stepSize = (info.maxAngle - info.minAngle) / (maxStep - minStep);
        device.lengthData.resize(lengthDataSize);

        connectedDevices.push_back(info);

        ARMARX_INFO << "Connected to " << device.ip << ", " << info.frame << ", "
                    << info.minAngle << ", " << info.maxAngle << ", " << info.stepSize;
    }

    task = new PeriodicTask<HokuyoLaserUnit>(this, &HokuyoLaserUnit::updateScanData, updatePeriod, false, "HokuyoLaserScanUpdate");
    task->setDelayWarningTolerance(100);
    task->start();
}


void HokuyoLaserUnit::onDisconnectComponent()
{
    if (task)
    {
        task->stop();
    }

    for (HokuyoLaserScanDevice& device : devices)
    {
        if (device.connected)
        {
            urg_close(&device.urg);
            device.connected = false;
        }
    }
}


void HokuyoLaserUnit::onExitComponent()
{

}

armarx::PropertyDefinitionsPtr HokuyoLaserUnit::createPropertyDefinitions()
{
    return armarx::PropertyDefinitionsPtr(new HokuyoLaserUnitPropertyDefinitions(
            getConfigIdentifier()));
}

std::string HokuyoLaserUnit::getReportTopicName(const Ice::Current& c) const
{
    return topicName;
}

LaserScannerInfoSeq HokuyoLaserUnit::getConnectedDevices(const Ice::Current& c) const
{
    return connectedDevices;
}

void HokuyoLaserUnit::updateScanData()
{
    LaserScan scan;
    TimestampVariantPtr now(new TimestampVariant(TimeUtil::GetTime()));
    for (HokuyoLaserScanDevice& device : devices)
    {
        if (device.connected)
        {
            int ret = urg_start_measurement(&device.urg, URG_DISTANCE, 1, 0);
            if (ret != 0)
            {
                ARMARX_WARNING << deactivateSpam(1) << "Could not start measurement for laser scanner (IP: "
                               << device.ip << ", Port: " << device.port << ", Error: " << ret << ")";
                continue;
            }

            int lengthDataSize = urg_get_distance(&device.urg, device.lengthData.data(), nullptr);
            if (lengthDataSize < 0)
            {
                ARMARX_WARNING << deactivateSpam(1) << "Could not get measurement for laser scanner (IP: "
                               << device.ip << ", Port: " << device.port << ", Error: " << lengthDataSize << ")";
                continue;
            }

            scan.clear();
            scan.reserve(lengthDataSize);
            for (int stepIndex = 0; stepIndex < lengthDataSize; ++stepIndex)
            {
                LaserScanStep step;
                long distance = device.lengthData[stepIndex];
                // TODO: Extract the min and max valid value for distance into parameters?
                if (distance >= 21 && distance <= 30000)
                {
                    step.angle = angleOffset + (float)urg_index2rad(&device.urg, stepIndex); // Convert steps to rad
                    step.distance = (float)distance; // Data is already in mm
                    scan.push_back(step);
                }
            }

            topic->reportSensorValues(device.ip, device.frame, scan, now);
        }
    }
}

