/*
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    RobotAPI::NJointControllerInterface
 * @author     Mirko Waechter ( mirko dot waechter at kit dot edu )
 * @date       2017
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#pragma once

#include <RobotAPI/interface/units/RobotUnit/NJointController.ice>
#include <RobotAPI/interface/core/Trajectory.ice>

module armarx
{
    module NJointTaskSpaceDMPControllerMode
    {
        enum CartesianSelection
        {
            ePosition = 7,
            eOrientation = 8,
            eAll = 15
        };
    };


    class NJointTaskSpaceDMPControllerConfig extends NJointControllerConfig
    {

        // dmp configuration
        float DMPKd = 20;
        int kernelSize = 100;
        double tau = 1;
        string dmpMode = "MinimumJerk";
        string dmpStyle = "Discrete";
        double dmpAmplitude = 1;

        double phaseL = 10;
        double phaseK = 10;
        double phaseDist0 = 50;
        double phaseDist1 = 10;
        double phaseKpPos = 1;
        double phaseKpOri = 0.1;
        double timeDuration = 10;
        double posToOriRatio = 100;

        // velocity controller configuration
        string nodeSetName = "";
        string tcpName = "";
        NJointTaskSpaceDMPControllerMode::CartesianSelection mode = NJointTaskSpaceDMPControllerMode::eAll;

        double maxLinearVel;
        double maxAngularVel;

        string debugName;

        float Kp_LinearVel;
        float Kd_LinearVel;
        float Kp_AngularVel;
        float Kd_AngularVel;
    };


    interface NJointTaskSpaceDMPControllerInterface extends NJointControllerInterface
    {
        void learnDMPFromFiles(Ice::StringSeq trajfiles);
        bool isFinished();
        void runDMP(Ice::DoubleSeq goals, double tau);

        void setSpeed(double times);
        void setViaPoints(double canVal, Ice::DoubleSeq point);
        void setGoals(Ice::DoubleSeq goals);

        double getCanVal();

        void setControllerTarget(float avoidJointLimitsKp, NJointTaskSpaceDMPControllerMode::CartesianSelection mode);
        void setTorqueKp(StringFloatDictionary torqueKp);
        void setNullspaceJointVelocities(StringFloatDictionary nullspaceJointVelocities);
    };

    class NJointCCDMPControllerConfig extends NJointControllerConfig
    {

        // dmp configuration
        float DMPKd = 20;
        int kernelSize = 100;
        double tau = 1;
        int dmpMode = 1;
        int dmpNum = 2;

        // phaseStop technique
        double phaseL = 10;
        double phaseK = 10;
        double phaseDist0 = 50;
        double phaseDist1 = 10;
        double phaseKpPos = 1;
        double phaseKpOri = 0.1;


        // misc
        Ice::DoubleSeq timeDurations;
        Ice::StringSeq dmpTypes;
        Ice::DoubleSeq amplitudes;


        double posToOriRatio = 100;

        // velocity controller configuration
        string nodeSetName = "";
        string tcpName = "";
        NJointTaskSpaceDMPControllerMode::CartesianSelection mode = NJointTaskSpaceDMPControllerMode::eAll;

    };

    interface NJointCCDMPControllerInterface extends NJointControllerInterface
    {
        void learnDMPFromFiles(int dmpId, Ice::StringSeq trajfiles);
        bool isFinished();
        void runDMP();

        void setTemporalFactor(int dmpId, double tau);
        void setViaPoints(int dmpId, double canVal, Ice::DoubleSeq point);
        void setGoals(int dmpId, Ice::DoubleSeq goals);

        void setControllerTarget(float avoidJointLimitsKp, NJointTaskSpaceDMPControllerMode::CartesianSelection mode);
        void setTorqueKp(StringFloatDictionary torqueKp);
        void setNullspaceJointVelocities(StringFloatDictionary nullspaceJointVelocities);
    };

    class NJointBimanualCCDMPControllerConfig extends NJointControllerConfig
    {

        // dmp configuration
        int kernelSize = 100;
        string dmpMode = "MinimumJerk";
        string dmpType = "Discrete";

        // phaseStop technique
        double phaseL = 10;
        double phaseK = 10;
        double phaseDist0 = 50;
        double phaseDist1 = 10;
        double phaseKpPos = 1;
        double phaseKpOri = 0.1;
        double posToOriRatio = 10;
        double timeDuration = 10;

        string defautLeader = "Left";

        Ice::FloatSeq leftDesiredJointValues;
        Ice::FloatSeq rightDesiredJointValues;

//        float KoriFollower = 1;
//        float KposFollower = 1;

        double maxLinearVel;
        double maxAngularVel;

        Ice::FloatSeq leftKpos;
        Ice::FloatSeq leftDpos;
        Ice::FloatSeq leftKori;
        Ice::FloatSeq leftDori;

        Ice::FloatSeq rightKpos;
        Ice::FloatSeq rightDpos;
        Ice::FloatSeq rightKori;
        Ice::FloatSeq rightDori;

        float knull;
        float dnull;

        float torqueLimit;

        float startReduceTorque;

//        Ice::FloatSeq Kpf;
//        Ice::FloatSeq Kif;
//        Ice::FloatSeq DesiredForce;

//        float BoxWidth;

//        float FilterTimeConstant;


    };

    interface NJointBimanualCCDMPControllerInterface extends NJointControllerInterface
    {
        void learnDMPFromFiles(string whichDMP, Ice::StringSeq trajfiles);
        bool isFinished();
        void runDMP(Ice::DoubleSeq leftGoals, Ice::DoubleSeq rightGoals);

        void setViaPoints(double canVal, Ice::DoubleSeq point);
        void setGoals(Ice::DoubleSeq goals);

        void changeLeader();
        double getVirtualTime();

        string getLeaderName();


    };

    class NJointTaskSpaceImpedanceDMPControllerConfig extends NJointControllerConfig
    {

        // dmp configuration
        int kernelSize = 100;
        string dmpMode = "MinimumJerk";
        string dmpType = "Discrete";
        double timeDuration;
        string nodeSetName;

        // phaseStop technique
        double phaseL = 10;
        double phaseK = 10;
        double phaseDist0 = 50;
        double phaseDist1 = 10;
        double posToOriRatio = 100;

        Ice::FloatSeq Kpos;
        Ice::FloatSeq Dpos;
        Ice::FloatSeq Kori;
        Ice::FloatSeq Dori;

        float Knull;
        float Dnull;



        bool useNullSpaceJointDMP;
        Ice::FloatSeq defaultNullSpaceJointValues;

        float torqueLimit;

    };

    interface NJointTaskSpaceImpedanceDMPControllerInterface extends NJointControllerInterface
    {
        void learnDMPFromFiles(Ice::StringSeq trajfiles);
        void learnJointDMPFromFiles(string jointTrajFile);

        bool isFinished();
        void runDMP(Ice::DoubleSeq goals);

        void setViaPoints(double canVal, Ice::DoubleSeq point);
        void setGoals(Ice::DoubleSeq goals);

        double getVirtualTime();

    };

};

