#include "NJointTSDMPController.h"

namespace armarx
{
    NJointControllerRegistration<NJointTSDMPController> registrationControllerNJointTSDMPController("NJointTSDMPController");

    NJointTSDMPController::NJointTSDMPController(const RobotUnitPtr&, const armarx::NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&)
    {
        useSynchronizedRtRobot();
        cfg = NJointTaskSpaceDMPControllerConfigPtr::dynamicCast(config);
        ARMARX_CHECK_EXPRESSION(!cfg->nodeSetName.empty());
        VirtualRobot::RobotNodeSetPtr rns = rtGetRobot()->getRobotNodeSet(cfg->nodeSetName);

        ARMARX_CHECK_EXPRESSION_W_HINT(rns, cfg->nodeSetName);
        for (size_t i = 0; i < rns->getSize(); ++i)
        {
            std::string jointName = rns->getNode(i)->getName();
            ControlTargetBase* ct = useControlTarget(jointName, ControlModes::Velocity1DoF);
            const SensorValueBase* sv = useSensorValue(jointName);
            targets.push_back(ct->asA<ControlTarget1DoFActuatorVelocity>());

            const SensorValue1DoFActuatorTorque* torqueSensor = sv->asA<SensorValue1DoFActuatorTorque>();
            const SensorValue1DoFActuatorVelocity* velocitySensor = sv->asA<SensorValue1DoFActuatorVelocity>();
            const SensorValue1DoFGravityTorque* gravityTorqueSensor = sv->asA<SensorValue1DoFGravityTorque>();
            const SensorValue1DoFActuatorPosition* positionSensor = sv->asA<SensorValue1DoFActuatorPosition>();
            if (!torqueSensor)
            {
                ARMARX_WARNING << "No Torque sensor available for " << jointName;
            }
            if (!gravityTorqueSensor)
            {
                ARMARX_WARNING << "No Gravity Torque sensor available for " << jointName;
            }

            torqueSensors.push_back(torqueSensor);
            gravityTorqueSensors.push_back(gravityTorqueSensor);
            velocitySensors.push_back(velocitySensor);
            positionSensors.push_back(positionSensor);
        };

        tcp = (cfg->tcpName.empty()) ? rns->getTCP() : rtGetRobot()->getRobotNode(cfg->tcpName);
        ARMARX_CHECK_EXPRESSION_W_HINT(tcp, cfg->tcpName);

        // set tcp controller
        tcpController.reset(new CartesianVelocityController(rns, tcp));
        nodeSetName = cfg->nodeSetName;
        torquePIDs.resize(tcpController->rns->getSize(), pidController());

        ik.reset(new VirtualRobot::DifferentialIK(rns, rns->getRobot()->getRootNode(), VirtualRobot::JacobiProvider::eSVDDamped));


        finished = false;
        TaskSpaceDMPControllerConfig taskSpaceDMPConfig;
        taskSpaceDMPConfig.motionTimeDuration = cfg->timeDuration;
        taskSpaceDMPConfig.DMPKernelSize = cfg->kernelSize;
        taskSpaceDMPConfig.DMPMode = cfg->dmpMode;
        taskSpaceDMPConfig.DMPStyle = cfg->dmpStyle;
        taskSpaceDMPConfig.DMPAmplitude = cfg->dmpAmplitude;
        taskSpaceDMPConfig.phaseStopParas.goDist = cfg->phaseDist0;
        taskSpaceDMPConfig.phaseStopParas.backDist = cfg->phaseDist1;
        taskSpaceDMPConfig.phaseStopParas.Kpos = cfg->phaseKpPos;
        taskSpaceDMPConfig.phaseStopParas.Dpos = 0;
        taskSpaceDMPConfig.phaseStopParas.Kori = cfg->phaseKpOri;
        taskSpaceDMPConfig.phaseStopParas.Dori = 0;
        taskSpaceDMPConfig.phaseStopParas.mm2radi = cfg->posToOriRatio;
        taskSpaceDMPConfig.phaseStopParas.maxValue = cfg->phaseL;
        taskSpaceDMPConfig.phaseStopParas.slop = cfg->phaseK;

        taskSpaceDMPController.reset(new TaskSpaceDMPController("default", taskSpaceDMPConfig, false));

        // initialize tcp position and orientation


        NJointTSDMPControllerSensorData initSensorData;
        initSensorData.deltaT = 0;
        initSensorData.currentTime = 0;
        initSensorData.currentPose.setZero();
        initSensorData.currentTwist.setZero();
        controllerSensorData.reinitAllBuffers(initSensorData);

        targetVels.resize(6);
        NJointTSDMPControllerControlData initData;
        initData.targetTSVel.resize(6);
        for (size_t i = 0; i < 6; ++i)
        {
            initData.targetTSVel(i) = 0;
            targetVels(i) = 0;
        }
        initData.nullspaceJointVelocities.resize(tcpController->rns->getSize(), 0);
        initData.torqueKp.resize(tcpController->rns->getSize(), 0);
        initData.torqueKd.resize(tcpController->rns->getSize(), 0);
        initData.mode = ModeFromIce(cfg->mode);
        reinitTripleBuffer(initData);

        debugName = cfg->debugName;

        KpF = cfg->Kp_LinearVel;
        KoF = cfg->Kp_AngularVel;
        DpF = cfg->Kd_LinearVel;
        DoF = cfg->Kd_AngularVel;

    }

    std::string NJointTSDMPController::getClassName(const Ice::Current&) const
    {
        return "NJointTSDMPController";
    }

    void NJointTSDMPController::controllerRun()
    {
        if (!controllerSensorData.updateReadBuffer() || !taskSpaceDMPController)
        {
            return;
        }
        double deltaT = controllerSensorData.getReadBuffer().deltaT;
        Eigen::Matrix4f currentPose = controllerSensorData.getReadBuffer().currentPose;
        Eigen::VectorXf currentTwist = controllerSensorData.getReadBuffer().currentTwist;

        taskSpaceDMPController->flow(deltaT, currentPose, currentTwist);

        if (taskSpaceDMPController->canVal < 1e-8)
        {
            finished = true;
        }
        targetVels = taskSpaceDMPController->getTargetVelocity();


        targetPose = taskSpaceDMPController->getTargetPoseMat();
        std::vector<double> targetState = taskSpaceDMPController->getTargetPose();
        debugOutputData.getWriteBuffer().latestTargetVelocities["x_vel"] = targetVels(0);
        debugOutputData.getWriteBuffer().latestTargetVelocities["y_vel"] = targetVels(1);
        debugOutputData.getWriteBuffer().latestTargetVelocities["z_vel"] = targetVels(2);
        debugOutputData.getWriteBuffer().latestTargetVelocities["roll_vel"] = targetVels(3);
        debugOutputData.getWriteBuffer().latestTargetVelocities["pitch_vel"] = targetVels(4);
        debugOutputData.getWriteBuffer().latestTargetVelocities["yaw_vel"] = targetVels(5);
        debugOutputData.getWriteBuffer().dmpTargets["dmp_x"] = targetState[0];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_y"] = targetState[1];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_z"] = targetState[2];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qw"] = targetState[3];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qx"] = targetState[4];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qy"] = targetState[5];
        debugOutputData.getWriteBuffer().dmpTargets["dmp_qz"] = targetState[6];
        debugOutputData.getWriteBuffer().currentPose["currentPose_x"] = currentPose(0, 3);
        debugOutputData.getWriteBuffer().currentPose["currentPose_y"] = currentPose(1, 3);
        debugOutputData.getWriteBuffer().currentPose["currentPose_z"] = currentPose(2, 3);

        VirtualRobot::MathTools::Quaternion currentQ = VirtualRobot::MathTools::eigen4f2quat(currentPose);
        debugOutputData.getWriteBuffer().currentPose["currentPose_qw"] = currentQ.w;
        debugOutputData.getWriteBuffer().currentPose["currentPose_qx"] = currentQ.x;
        debugOutputData.getWriteBuffer().currentPose["currentPose_qy"] = currentQ.y;
        debugOutputData.getWriteBuffer().currentPose["currentPose_qz"] = currentQ.z;
        debugOutputData.getWriteBuffer().currentCanVal = taskSpaceDMPController->debugData.canVal;
        debugOutputData.getWriteBuffer().mpcFactor =  taskSpaceDMPController->debugData.mpcFactor;
        debugOutputData.getWriteBuffer().error = taskSpaceDMPController->debugData.poseError;
        debugOutputData.getWriteBuffer().posError = taskSpaceDMPController->debugData.posiError;
        debugOutputData.getWriteBuffer().oriError = taskSpaceDMPController->debugData.oriError;
        debugOutputData.getWriteBuffer().deltaT = deltaT;

        debugOutputData.commitWrite();

        getWriterControlStruct().targetTSVel = targetVels;
        getWriterControlStruct().targetPose = targetPose;

        writeControlStruct();
    }


    void NJointTSDMPController::rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration)
    {
        Eigen::Matrix4f currentPose = tcp->getPoseInRootFrame();
        double deltaT = timeSinceLastIteration.toSecondsDouble();

        Eigen::MatrixXf jacobi = ik->getJacobianMatrix(tcp, VirtualRobot::IKSolver::CartesianSelection::All);

        Eigen::VectorXf qvel;
        qvel.resize(velocitySensors.size());
        for (size_t i = 0; i < velocitySensors.size(); ++i)
        {
            qvel(i) = velocitySensors[i]->velocity;
        }

        Eigen::VectorXf tcptwist = jacobi * qvel;

        controllerSensorData.getWriteBuffer().currentPose = currentPose;
        controllerSensorData.getWriteBuffer().currentTwist = tcptwist;
        controllerSensorData.getWriteBuffer().deltaT = deltaT;
        controllerSensorData.getWriteBuffer().currentTime += deltaT;
        controllerSensorData.commitWrite();


        Eigen::VectorXf targetVel = rtGetControlStruct().targetTSVel;
        Eigen::Matrix4f targetPose = rtGetControlStruct().targetPose;


        Eigen::Matrix3f diffMat = targetPose.block<3, 3>(0, 0) * currentPose.block<3, 3>(0, 0).inverse();
        Eigen::Vector3f errorRPY = VirtualRobot::MathTools::eigen3f2rpy(diffMat);


        Eigen::VectorXf rtTargetVel;
        rtTargetVel.resize(6);
        rtTargetVel.block<3, 1>(0, 0) = KpF * (targetPose.block<3, 1>(0, 3) - currentPose.block<3, 1>(0, 3)) + DpF * (targetVel.block<3, 1>(0, 0) - tcptwist.block<3, 1>(0, 0));
        rtTargetVel.block<3, 1>(3, 0) = KoF * errorRPY + DoF * (targetVel.block<3, 1>(3, 0) - tcptwist.block<3, 1>(3, 0));
        //        rtTargetVel = targetVel;


        float normLinearVelocity = rtTargetVel.block<3, 1>(0, 0).norm();
        if (normLinearVelocity > cfg->maxLinearVel)
        {
            rtTargetVel.block<3, 1>(0, 0) = cfg->maxLinearVel * rtTargetVel.block<3, 1>(0, 0) / normLinearVelocity;
        }

        float normAngularVelocity = rtTargetVel.block<3, 1>(3, 0).norm();
        if (normAngularVelocity > cfg->maxAngularVel)
        {
            rtTargetVel.block<3, 1>(3, 0) = cfg->maxAngularVel * rtTargetVel.block<3, 1>(3, 0) / normAngularVelocity;
        }


        // cartesian vel controller
        Eigen::VectorXf x;
        auto mode = rtGetControlStruct().mode;

        if (mode == VirtualRobot::IKSolver::All)
        {
            x.resize(6);
            for (size_t i = 0; i < 6; i++)
            {
                x(i) = rtTargetVel(i);
            }
        }
        else if (mode == VirtualRobot::IKSolver::Position)
        {
            x.resize(3);
            for (size_t i = 0; i < 3; i++)
            {
                x(i) = rtTargetVel(i);
            }

        }
        else if (mode == VirtualRobot::IKSolver::Orientation)
        {
            x.resize(3);
            for (size_t i = 0; i < 3; i++)
            {
                x(i) = rtTargetVel(i + 3);
            }
        }
        else
        {

            // No mode has been set
            return;
        }

        Eigen::VectorXf jnv = Eigen::VectorXf::Zero(tcpController->rns->getSize());
        float jointLimitAvoidanceKp = rtGetControlStruct().avoidJointLimitsKp;
        if (jointLimitAvoidanceKp > 0)
        {
            jnv += jointLimitAvoidanceKp * tcpController->calculateJointLimitAvoidance();
        }
        for (size_t i = 0; i < tcpController->rns->getSize(); i++)
        {
            jnv(i) += rtGetControlStruct().nullspaceJointVelocities.at(i);
            if (torqueSensors.at(i) && gravityTorqueSensors.at(i) && rtGetControlStruct().torqueKp.at(i) != 0)
            {
                torquePIDs.at(i).Kp = rtGetControlStruct().torqueKp.at(i);
                torquePIDs.at(i).Kd = rtGetControlStruct().torqueKd.at(i);
                jnv(i) += torquePIDs.at(i).update(deltaT, torqueSensors.at(i)->torque - gravityTorqueSensors.at(i)->gravityTorque);
            }
            else
            {
                torquePIDs.at(i).lastError = 0;
            }
        }

        Eigen::VectorXf jointTargetVelocities = tcpController->calculate(x, jnv, mode);
        for (size_t i = 0; i < targets.size(); ++i)
        {
            targets.at(i)->velocity = jointTargetVelocities(i);
        }


    }


    void NJointTSDMPController::learnDMPFromFiles(const Ice::StringSeq& fileNames, const Ice::Current&)
    {
        taskSpaceDMPController->learnDMPFromFiles(fileNames);

        ARMARX_INFO << "Learned DMP ... ";
    }

    void NJointTSDMPController::setSpeed(Ice::Double times, const Ice::Current&)
    {
        LockGuardType guard(controllerMutex);
        taskSpaceDMPController->setSpeed(times);
    }

    void NJointTSDMPController::setViaPoints(Ice::Double u, const Ice::DoubleSeq& viapoint, const Ice::Current&)
    {
        LockGuardType guard(controllerMutex);
        ARMARX_INFO << "setting via points ";
        taskSpaceDMPController->setViaPose(u, viapoint);
    }


    void NJointTSDMPController::setGoals(const Ice::DoubleSeq& goals, const Ice::Current& ice)
    {
        taskSpaceDMPController->setGoalPoseVec(goals);
    }

    void NJointTSDMPController::setControllerTarget(Ice::Float avoidJointLimitsKp, NJointTaskSpaceDMPControllerMode::CartesianSelection mode, const Ice::Current&)
    {
        LockGuardType guard {controlDataMutex};
        getWriterControlStruct().avoidJointLimitsKp = avoidJointLimitsKp;
        getWriterControlStruct().mode = ModeFromIce(mode);
        writeControlStruct();
    }

    VirtualRobot::IKSolver::CartesianSelection NJointTSDMPController::ModeFromIce(const NJointTaskSpaceDMPControllerMode::CartesianSelection mode)
    {
        if (mode == NJointTaskSpaceDMPControllerMode::CartesianSelection::ePosition)
        {
            return VirtualRobot::IKSolver::CartesianSelection::Position;
        }
        if (mode == NJointTaskSpaceDMPControllerMode::CartesianSelection::eOrientation)
        {
            return VirtualRobot::IKSolver::CartesianSelection::Orientation;
        }
        if (mode == NJointTaskSpaceDMPControllerMode::CartesianSelection::eAll)
        {
            return VirtualRobot::IKSolver::CartesianSelection::All;
        }
        ARMARX_ERROR_S << "invalid mode " << mode;
        return (VirtualRobot::IKSolver::CartesianSelection)0;
    }


    void NJointTSDMPController::setTorqueKp(const StringFloatDictionary& torqueKp, const Ice::Current&)
    {
        LockGuardType guard {controlDataMutex};
        for (size_t i = 0; i < tcpController->rns->getSize(); i++)
        {
            getWriterControlStruct().torqueKp.at(i) = torqueKp.at(tcpController->rns->getNode(i)->getName());
        }
        writeControlStruct();
    }

    void NJointTSDMPController::setNullspaceJointVelocities(const StringFloatDictionary& nullspaceJointVelocities, const Ice::Current&)
    {
        LockGuardType guard {controlDataMutex};
        for (size_t i = 0; i < tcpController->rns->getSize(); i++)
        {
            getWriterControlStruct().nullspaceJointVelocities.at(i) = nullspaceJointVelocities.at(tcpController->rns->getNode(i)->getName());
        }
        writeControlStruct();
    }

    void NJointTSDMPController::runDMP(const Ice::DoubleSeq&  goals, double tau, const Ice::Current&)
    {
        Eigen::Matrix4f pose = tcp->getPoseInRootFrame();
        taskSpaceDMPController->prepareExecution(taskSpaceDMPController->eigen4f2vec(pose), goals);
        finished = false;

        ARMARX_INFO << "run DMP";
        controllerTask->start();

    }


    void NJointTSDMPController::rtPreActivateController()
    {
    }

    void NJointTSDMPController::rtPostDeactivateController()
    {

    }

    void NJointTSDMPController::onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx& debugObs)
    {
        std::string datafieldName = debugName;
        StringVariantBaseMap datafields;
        auto values = debugOutputData.getUpToDateReadBuffer().latestTargetVelocities;
        for (auto& pair : values)
        {
            datafieldName = pair.first  + "_" + debugName;
            datafields[datafieldName] = new Variant(pair.second);
        }

        auto dmpTargets = debugOutputData.getUpToDateReadBuffer().dmpTargets;
        for (auto& pair : dmpTargets)
        {
            datafieldName = pair.first  + "_" + debugName;
            datafields[datafieldName] = new Variant(pair.second);
        }

        auto currentPose = debugOutputData.getUpToDateReadBuffer().currentPose;
        for (auto& pair : currentPose)
        {
            datafieldName = pair.first + "_" + debugName;
            datafields[datafieldName] = new Variant(pair.second);
        }

        datafieldName = "canVal_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().currentCanVal);
        datafieldName = "mpcFactor_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().mpcFactor);
        datafieldName = "error_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().error);
        datafieldName = "posError_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().posError);
        datafieldName = "oriError_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().oriError);
        datafieldName = "deltaT_" + debugName;
        datafields[datafieldName] = new Variant(debugOutputData.getUpToDateReadBuffer().deltaT);
        datafieldName = "DMPController_" + debugName;

        debugObs->setDebugChannel(datafieldName, datafields);
    }

    void NJointTSDMPController::onInitComponent()
    {
        ARMARX_INFO << "init ...";
        controllerTask = new PeriodicTask<NJointTSDMPController>(this, &NJointTSDMPController::controllerRun, 0.3);
    }

    void NJointTSDMPController::onDisconnectComponent()
    {
        ARMARX_INFO << "stopped ...";
        controllerTask->stop();
    }



}
