
#pragma once

#include <RobotAPI/components/units/RobotUnit/NJointControllers/NJointController.h>
#include <VirtualRobot/Robot.h>
#include <RobotAPI/components/units/RobotUnit/RobotUnit.h>
#include <RobotAPI/components/units/RobotUnit/ControlTargets/ControlTarget1DoFActuator.h>
#include <RobotAPI/components/units/RobotUnit/SensorValues/SensorValue1DoFActuator.h>
#include <VirtualRobot/IK/DifferentialIK.h>
#include <dmp/representation/dmp/umitsmp.h>
#include <RobotAPI/interface/units/RobotUnit/NJointTaskSpaceDMPController.h>
#include <RobotAPI/libraries/core/CartesianVelocityController.h>

#include <RobotAPI/libraries/DMPController/TaskSpaceDMPController.h>

using namespace DMP;
namespace armarx
{


    TYPEDEF_PTRS_HANDLE(NJointTSDMPController);
    TYPEDEF_PTRS_HANDLE(NJointTSDMPControllerControlData);

    typedef std::pair<double, DVec > ViaPoint;
    typedef std::vector<ViaPoint > ViaPointsSet;
    class NJointTSDMPControllerControlData
    {
    public:
        Eigen::VectorXf targetTSVel;
        Eigen::Matrix4f targetPose;
        // cartesian velocity control data
        std::vector<float> nullspaceJointVelocities;
        float avoidJointLimitsKp = 0;
        std::vector<float> torqueKp;
        std::vector<float> torqueKd;
        VirtualRobot::IKSolver::CartesianSelection mode = VirtualRobot::IKSolver::All;
    };

    class pidController
    {
    public:
        float Kp = 0, Kd = 0;
        float lastError = 0;
        float update(float dt, float error)
        {
            float derivative = (error - lastError) / dt;
            float retVal = Kp * error + Kd * derivative;
            lastError = error;
            return retVal;
        }
    };

    /**
     * @brief The NJointTSDMPController class
     * @ingroup Library-RobotUnit-NJointControllers
     */
    class NJointTSDMPController :
        public NJointControllerWithTripleBuffer<NJointTSDMPControllerControlData>,
        public NJointTaskSpaceDMPControllerInterface
    {
    public:
        using ConfigPtrT = NJointTaskSpaceDMPControllerConfigPtr;
        NJointTSDMPController(const RobotUnitPtr&, const NJointControllerConfigPtr& config, const VirtualRobot::RobotPtr&);

        // NJointControllerInterface interface
        std::string getClassName(const Ice::Current&) const;

        // NJointController interface

        void rtRun(const IceUtil::Time& sensorValuesTimestamp, const IceUtil::Time& timeSinceLastIteration);

        // NJointTSDMPControllerInterface interface
        void learnDMPFromFiles(const Ice::StringSeq& fileNames, const Ice::Current&);
        bool isFinished(const Ice::Current&)
        {
            return finished;
        }

        void runDMP(const Ice::DoubleSeq& goals, Ice::Double tau, const Ice::Current&);
        void setSpeed(Ice::Double times, const Ice::Current&);
        void setViaPoints(Ice::Double u, const Ice::DoubleSeq& viapoint, const Ice::Current&);

        void setGoals(const Ice::DoubleSeq& goals, const Ice::Current&);

        void setControllerTarget(Ice::Float avoidJointLimitsKp, NJointTaskSpaceDMPControllerMode::CartesianSelection mode, const Ice::Current&);
        void setTorqueKp(const StringFloatDictionary& torqueKp, const Ice::Current&);
        void setNullspaceJointVelocities(const StringFloatDictionary& nullspaceJointVelocities, const Ice::Current&);

        double getCanVal(const Ice::Current&)
        {
            return taskSpaceDMPController->canVal;
        }

    protected:
        void rtPreActivateController() override;
        void rtPostDeactivateController() override;
        VirtualRobot::IKSolver::CartesianSelection ModeFromIce(const NJointTaskSpaceDMPControllerMode::CartesianSelection mode);
        virtual void onPublish(const SensorAndControl&, const DebugDrawerInterfacePrx&, const DebugObserverInterfacePrx&);

        void onInitComponent();
        void onDisconnectComponent();
        void controllerRun();

    private:
        struct DebugBufferData
        {
            StringFloatDictionary latestTargetVelocities;
            StringFloatDictionary dmpTargets;
            StringFloatDictionary currentPose;
            double currentCanVal;
            double mpcFactor;
            double error;
            double phaseStop;
            double posError;
            double oriError;
            double deltaT;
        };

        TripleBuffer<DebugBufferData> debugOutputData;

        struct NJointTSDMPControllerSensorData
        {
            double currentTime;
            double deltaT;
            Eigen::Matrix4f currentPose;
            Eigen::VectorXf currentTwist;
        };
        TripleBuffer<NJointTSDMPControllerSensorData> controllerSensorData;


        std::vector<const SensorValue1DoFActuatorTorque*> torqueSensors;
        std::vector<const SensorValue1DoFGravityTorque*> gravityTorqueSensors;
        std::vector<ControlTarget1DoFActuatorVelocity*> targets;
        std::vector<const SensorValue1DoFActuatorVelocity*> velocitySensors;
        std::vector<const SensorValue1DoFActuatorPosition*> positionSensors;


        TaskSpaceDMPControllerPtr taskSpaceDMPController;

        // velocity ik controller parameters
        std::vector<pidController> torquePIDs;
        CartesianVelocityControllerPtr tcpController;
        std::string nodeSetName;

        // dmp parameters
        bool finished;

        VirtualRobot::DifferentialIKPtr ik;
        VirtualRobot::RobotNodePtr tcp;
        Eigen::VectorXf targetVels;
        Eigen::Matrix4f targetPose;

        NJointTaskSpaceDMPControllerConfigPtr cfg;
        mutable MutexType controllerMutex;
        PeriodicTask<NJointTSDMPController>::pointer_type controllerTask;


        std::string debugName;

        float KpF;
        float DpF;
        float KoF;
        float DoF;
    };

} // namespace armarx

