/*
 * This file is part of ArmarX.
 *
 * Copyright (C) 2012-2016, High Performance Humanoid Technologies (H2T),
 * Karlsruhe Institute of Technology (KIT), all rights reserved.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @author     Simon Ottenhaus (simon dot ottenhaus at kit dot edu)
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include "RobotNameHelper.h"
#include <ArmarXCore/core/util/StringHelpers.h>

using namespace armarx;

const std::string RobotNameHelper::LocationLeft = "Left";
const std::string RobotNameHelper::LocationRight = "Right";

RobotNameHelper::RobotNameHelper(const RobotInfoNodePtr& robotInfo, const StatechartProfilePtr& profile)
    : root(Node(robotInfo))
{
    StatechartProfilePtr p = profile;
    while (p && !p->isRoot())
    {
        profiles.emplace_back(p->getName());
        p = p->getParent();
    }
    profiles.emplace_back(""); // for matching the default/root

}

std::string RobotNameHelper::select(const std::string& path) const
{
    Node node = root;
    for (const std::string& p : Split(path, "/"))
    {
        node = node.select(p, profiles);
    }
    if (!node.isValid())
    {
        throw std::runtime_error("RobotNameHelper::select: path " + path + " not found");
    }
    return node.value();
}

RobotNameHelperPtr RobotNameHelper::Create(const RobotInfoNodePtr& robotInfo, const StatechartProfilePtr& profile)
{
    return RobotNameHelperPtr(new RobotNameHelper(robotInfo, profile));
}

RobotNameHelper::Arm RobotNameHelper::getArm(const std::string& side)
{
    return Arm(shared_from_this(), side);
}

RobotNameHelper::RobotArm RobotNameHelper::getRobotArm(const std::string& side, const VirtualRobot::RobotPtr& robot)
{
    return RobotArm(Arm(shared_from_this(), side), robot);
}

RobotNameHelper::Node::Node(const RobotInfoNodePtr& robotInfo)
    : robotInfo(robotInfo)
{

}

std::string RobotNameHelper::Node::value()
{
    checkValid();
    return robotInfo->value;
}

RobotNameHelper::Node RobotNameHelper::Node::select(const std::string& name, const std::vector<std::string>& profiles)
{
    if (!isValid())
    {
        return Node(nullptr);
    }
    std::map<std::string, RobotInfoNodePtr> matches;
    for (const RobotInfoNodePtr& child : robotInfo->children)
    {
        if (child->name == name)
        {
            matches[child->profile] = child;
        }
    }
    for (const std::string& p : profiles)
    {
        if (matches.count(p))
        {
            return matches.at(p);
        }
    }
    return Node(nullptr);
}

bool RobotNameHelper::Node::isValid()
{
    return robotInfo ? true : false;
}

void RobotNameHelper::Node::checkValid()
{
    if (!isValid())
    {
        throw std::runtime_error("RobotNameHelper::Node nullptr exception");
    }
}


std::string RobotNameHelper::Arm::getKinematicChain() const
{
    return select("KinematicChain");
}

std::string RobotNameHelper::Arm::getTorsoKinematicChain() const
{
    return select("TorsoKinematicChain");
}

std::string RobotNameHelper::Arm::getTCP() const
{
    return select("TCP");
}

std::string RobotNameHelper::Arm::getForceTorqueSensor() const
{
    return select("ForceTorqueSensor");
}

std::string RobotNameHelper::Arm::getEndEffector() const
{
    return select("EndEffector");
}

std::string RobotNameHelper::Arm::getMemoryHandName() const
{
    return select("MemoryHandName");
}

std::string RobotNameHelper::Arm::getHandControllerName() const
{
    return select("HandControllerName");
}

RobotNameHelper::RobotArm RobotNameHelper::Arm::addRobot(const VirtualRobot::RobotPtr& robot) const
{
    return RobotArm(*this, robot);
}

RobotNameHelper::Arm::Arm(const std::shared_ptr<RobotNameHelper>& rnh, const std::string& side)
    : rnh(rnh), side(side)
{

}

std::string RobotNameHelper::Arm::select(const std::string& path) const
{
    return rnh->select(side + "Arm/" + path);
}

VirtualRobot::RobotNodeSetPtr RobotNameHelper::RobotArm::getKinematicChain() const
{
    return robot->getRobotNodeSet(arm.getKinematicChain());
}

VirtualRobot::RobotNodeSetPtr RobotNameHelper::RobotArm::getTorsoKinematicChain() const
{
    return robot->getRobotNodeSet(arm.getTorsoKinematicChain());
}

VirtualRobot::RobotNodePtr RobotNameHelper::RobotArm::getTCP() const
{
    return robot->getRobotNode(arm.getTCP());
}

RobotNameHelper::RobotArm::RobotArm(const Arm& arm, const VirtualRobot::RobotPtr& robot)
    : arm(arm), robot(robot)
{

}

